# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys


class ReportingReader:
    """
    Utility wrapper used for reporting on the progress of reading a file.

    The ReportingReader can be used as a normal file-like object. The main
    difference with a general file-like object is that each time "read" is
    called the read progress is printed to the console.

    For example, to report the progress of uploading a file you can use:

        with ReportingReader(massive_file_path) as local_file:
            requests.post(url, data=local_file)

    This will use the default streaming uploads capability of requests to
    upload a file, but displays the upload progress on the terminal.
    """

    def __init__(self, file_path, mode="rb", progress_prefix=""):
        self._previous_status = ""
        self._progress_prefix = progress_prefix
        self._file_size = os.path.getsize(file_path)
        self._reader = open(file_path, mode)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        ReportingReader.write_progress("")
        self._reader.close()

    @staticmethod
    def write_progress(string):
        if not sys.stderr.isatty:
            return
        sys.stderr.write("\r" + string)

    def read(self, size=-1):
        progress = self._reader.tell() / self._file_size * 100
        status = f"{self._progress_prefix}{progress:.0f}%"
        if status != self._previous_status:
            ReportingReader.write_progress(status)
            self._previous_status = status
        return self._reader.read(size)



# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import csv
import yaml
from datetime import datetime
from functools import total_ordering

import prettytable
from six import StringIO
from six.moves import zip

from .datetimefunctions import format_datetime, get_datetime_ago


@total_ordering
class NoneWrapper:
    """
    Object of any data type can be compared to the instance of this class.
    No error is raised with any comparison operators.
    Needed to convert None into its instance to sort table with unexpected, mixed data types.
    """
    def __init__(self):
        pass

    def __eq__(self, other):
        if isinstance(other, NoneWrapper):
            return True
        else:
            return False

    def __gt__(self, other):
        if isinstance(other, NoneWrapper):
            return False
        else:
            return True

    def __str__(self):
        return "?"

    def __repr__(self):
        return "?"


@total_ordering
class SSHAlias:
    """
    Small helper class for our SSH alias column.

    We sort the SSH aliases based on the index of the alias, not on the prefix. E.g. os1 and os2
    are sorted as 1 and 2. For some clusters there won't be an SSH alias, for example while the
    cluster is being created (there are other cases as well). We want to display those as "?",
    but we also need to be able to sort on the column. So for sorting we treat "?" as 0.
    """
    def __init__(self, alias, prefix=""):
        self._alias = alias

        if isinstance(alias, str):
            self._index = alias.replace(prefix, "")
        elif isinstance(alias, int):
            self._index = alias
        else:
            raise TypeError(f"alias of type {type(alias)} not supported")

    @staticmethod
    def _index_to_int(index):
        try:
            return int(index)
        except ValueError:
            return 0

    def __eq__(self, other):
        return self._index_to_int(self._index) == self._index_to_int(other._index)

    def __gt__(self, other):
        return self._index_to_int(self._index) > self._index_to_int(other._index)

    def __str__(self):
        return str(self._alias)

    def __repr__(self):
        return str(self._alias)


class SortableData(object):
    """
    Utility class for dealing with sortable data.

    :param all_headers: Headers of all columns in following format
        [
            (column_id, column_name),
            ....
        ]
    :param requested_headers: a list of ids of the columns to be displayed.
        If empty. All columns are displayed.
    :param rows: Rows of data
    """

    def __init__(self, all_headers, requested_headers, rows):
        self.all_headers = all_headers
        self.requested_headers = [
            col_data for col_data in self.all_headers
            if not requested_headers or col_data[0] in requested_headers
        ]
        self.rows = rows

    def sort(self, *sorting_columns):
        columns = [header[0] for header in self.all_headers]
        filtered_sorting_columns = [s for s in sorting_columns if s in columns]

        # pick first one by default
        if not filtered_sorting_columns:
            filtered_sorting_columns = [columns[0]]

        column_ids = [elt[0] for elt in self.all_headers]

        self.column_indices = [
            column_ids.index(sorting_column) for sorting_column in filtered_sorting_columns
        ]

        self.rows = [[NoneWrapper() if col is None else col for col in row] for row in self.rows]

        self.sorted_rows = sorted(
            self.rows, key=lambda x: [
                x[column_index] for column_index in self.column_indices
            ]
        )

        self.filter(self.sorted_rows)
        self.sorted_rows = self._format_datetime(self.sorted_rows)

    def filter(self, all_rows):
        requested_indices = []
        filtered_rows = []
        filtered_columns = []

        for column_data in self.all_headers:
            if column_data in self.requested_headers and column_data not in filtered_columns:
                filtered_columns.append(column_data)
                requested_indices += [self.all_headers.index(column_data)]

        for row in all_rows:
            filtered_rows.append([row[x] for x in requested_indices])

        self.filtered_columns_data, self.sorted_rows = filtered_columns, filtered_rows

    def output(self, output_format):
        if output_format == "table":
            return self.make_pretty_table(self.filtered_columns_data, self.sorted_rows)
        elif output_format == "csv":
            return self.make_csv_table(self.filtered_columns_data, self.sorted_rows)
        elif output_format == "value":
            return self.make_value_output(self.filtered_columns_data, self.sorted_rows)
        elif output_format == "yaml":
            return self.make_yaml_output(self.filtered_columns_data, self.sorted_rows)
        assert False, "output_format was of unrecognized format: %s" % (output_format)

    @staticmethod
    def make_pretty_table(columns, data):
        columns_names = [column[1] for column in columns]
        table = prettytable.PrettyTable(columns_names)
        table.align = "l"
        for row in data:
            table.add_row(row)
        return table

    @staticmethod
    def make_csv_table(columns, data):
        columns_names = [column[1] for column in columns]
        csv_output = StringIO()
        csv_writer = csv.writer(csv_output)
        csv_writer.writerow(columns_names)
        csv_writer.writerows(data)

        return csv_output.getvalue()

    @staticmethod
    def make_value_output(columns, data):
        def escape(cell):
            cell = str(cell)
            if " " in cell:
                cell = '"' + cell.replace('"', '\\"') + '"'
            elif cell == "":
                cell = '""'
            return cell

        def make_line(row):
            return " ".join([escape(cell) for cell in row])

        lines = [make_line(row) for row in data]
        return "\n".join(lines)

    @staticmethod
    def make_yaml_output(columns, data):
        results = []

        for row in data:
            results.append({col[0]: str(cell) for col, cell in zip(columns, row)})

        output = {"results": results}
        return yaml.safe_dump(output, default_flow_style=False)

    @staticmethod
    def _format_datetime(rows):
        for i, row in enumerate(rows):
            for j, cell in enumerate(row):
                if isinstance(cell, datetime):
                    rows[i][j] = "{created} ({ago})".format(
                        created=format_datetime(rows[i][j]),
                        ago=get_datetime_ago(rows[i][j])
                    )
        return rows



# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from six.moves import range

from clusterondemand.exceptions import CODException


def expand_vlan_ids(vlan_ids):
    """Translate [[20, 22], 10] to [10, 20, 21, 22]."""
    ret = []
    for id_range in vlan_ids:
        len_ = len(id_range)
        if len_ == 1:
            ret.append(id_range[0])
        elif len_ == 2:
            ret.extend(range(id_range[0], id_range[1] + 1))
        else:
            raise Exception("Bad value for vlan IDs: %s" % str(id_range))
    return sorted(ret)


def vlan_id_parser(string):
    try:
        return [int(string)]
    except ValueError:
        pass

    try:
        start, end = string.split("-")
        return [int(start), int(end)]
    except ValueError as e:
        raise CODException(
            "Invalid vlan ID format. Should be in '[<start ID>-<end ID> | ID] "
            "[[<start ID>-<end ID> | ID]]' format. Value: %s" % string,
            caused_by=e
        )


def vlan_id_serializer(value):
    if len(value) == 1:
        return str(value[0])
    elif len(value) == 2:
        return str(value[0]) + "-" + str(value[1])
    else:
        # values should either be individual ID, or start/end pair
        assert False, "Bad value %s for vlan ID." % str(value)



# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

from copy import deepcopy
from logging import getLogger

from six.moves.urllib.parse import urljoin

log = getLogger("cluster-on-demand")


class ImageDescriptionVersionMismatchError(Exception):
    """Version of the description object is not supported."""

    def __init__(self, version):
        super(ImageDescriptionVersionMismatchError, self).__init__(
            "Version %s of image description is not supported", version
        )


class Image(object):
    """Image description object.

    It can be constructed from the description returned from Repository.
    """

    @classmethod
    def make_from_descripton(cls, descr, baseurl):
        """Create image object from descr object."""
        if descr["version"] != 1:
            raise ImageDescriptionVersionMismatchError(descr["version"])

        if baseurl is not None:
            descr = deepcopy(descr)
            descr["url"] = urljoin(baseurl, descr["url"])
        return cls(descr)

    @classmethod
    def make_from_descripton_noraise(cls, descr, baseurl):
        """Create image object from descr object."""
        try:
            return cls.make_from_descripton(descr, baseurl)
        except Exception as e:
            log.warning("Could not parse image description %s: %s", descr, e)

    @classmethod
    def bootstrap_image(cls,
                        id,
                        url,
                        tags,
                        package_groups,
                        bcm_version,
                        bcm_release,
                        image_type,
                        os,
                        revision,
                        cmd_revision,
                        created_at,
                        cloud_type):
        return cls.make_from_descripton({
            "version": 1,
            "id": id,
            "url": url,
            "tags": tags,
            "package_groups": package_groups,
            "bcm_version": bcm_version,
            "bcm_release": bcm_release,
            "image_type": image_type,
            "os": os,
            "revision": revision,
            "cmd_revision": cmd_revision,
            "created_at": created_at,
            "cloud_type": cloud_type,
        }, baseurl=None)

    def uncompressed_image_info_missing(self):
        return self._description.get("uncompressed_size") is None

    def image_info_missing(self):
        return self._description.get("size") is None or \
            self._description.get("md5sum") is None

    def set_uncompressed_image_info(self, uncompressed_size):
        self._description["uncompressed_size"] = uncompressed_size

    def set_image_info(self, size, md5sum):
        self._description["size"] = size
        self._description["md5sum"] = md5sum

    def __init__(self, description):
        self._description = description

    def raw_description(self):
        return self._description

    def id(self):
        """Return unique string across the repo."""
        return self._description["id"]

    def url(self):
        """Return URL of the image file."""
        return self._description["url"]

    def tags(self):
        """Return tags of the image."""
        return self._description["tags"]

    def md5sum(self):
        """Return MD5 checksum of the image file."""
        return self._description["md5sum"]

    def size(self):
        """Size of the image file in bytes."""
        return self._description["size"]

    def uncompressed_size(self):
        return self._description["uncompressed_size"]

    def package_groups(self):
        """Return package groups preinstalled to the image."""
        return self._description["package_groups"]

    def bcm_version(self):
        """Bright version installed to the image."""
        return self._description["bcm_version"]

    def bcm_release(self):
        """Bright release installed to the image."""
        return self._description.get("bcm_release", 0)

    def image_type(self):
        """Type of bcm image."""
        return self._description["image_type"]

    def os(self):
        """Operating system of the image."""
        return self._description["os"]

    def revision(self):
        """Image revision."""
        return self._description["revision"]

    def created_at(self):
        """Image creation timestamp."""
        return self._description["created_at"]

    def cmd_revision(self):
        """Return CMDaemon revision installed in the image."""
        return self._description["cmd_revision"]

    def cloud_type(self):
        """Cloud type this image was created for."""
        return self._description["cloud_type"]

    def bcm_api_hash(self):
        """CMD API version hash, used by API client such as pythoncm."""
        return self._description["bcm_api_hash"]

    def completed(self):
        """Object is consistent and contains all necessary information."""
        def _check_key(key):
            return self._description.get(key) is not None

        keys = [
            "id",
            "url",
            "tags",
            "md5sum",
            "size",
            "uncompressed_size",
            "package_groups",
            "bcm_version",
            "image_type",
            "os",
            "revision",
            "created_at",
            "cmd_revision",
            "version",
            "cloud_type",
        ]
        for key in keys:
            if not _check_key(key):
                return False

        return True



# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import netaddr

from .exceptions import UserReportableException


def nullable_cidr(string):
    """A parser for cidr strings which allows None values as well."""
    if string.lower() == "none":
        return None
    return cidr(string)


def cidr(string):
    """A parser for cidr strings."""
    try:
        network = netaddr.IPNetwork(string)
        if str(network.cidr) != str(network):
            raise UserReportableException("'%s' is not a valid CIDR: the host section is not empty" % (string))
        return network
    except netaddr.core.AddrFormatError:
        raise UserReportableException("'%s' is not a valid CIDR" % (string))


def must_be_within_cidr(parent_cidr_name):
    """Validation that raises an error when the CIDR value of the parameter is not contained within the parent CIDR"""
    def wrapper(parameter, configuration):
        parent = configuration.get_item_for_key(parent_cidr_name).parameter
        subnet_cidr = configuration[parameter.key]
        parent_cidr = configuration[parent.key]

        if subnet_cidr is None:
            raise UserReportableException("%s is not set" % (parameter.key))
        if parent_cidr is None:
            raise UserReportableException("%s is not set" % (parent.key))

        if subnet_cidr.network < parent_cidr.network or subnet_cidr.broadcast > parent_cidr.broadcast:
            raise UserReportableException(
                "%s=%s not within %s=%s. Use %s and/or %s to change that." % (
                    parameter.key, subnet_cidr,
                    parent.key, parent_cidr,
                    parent.default_flag, parameter.default_flag
                )
            )

    return wrapper



# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from clusterondemandconfig import ConfigLoadError

from .clustercreate import generate_random_cluster_password

log = logging.getLogger("cluster-on-demand")


def is_ssh_key_specified(configuration):
    # Depending on the COD implementation there are two different types of ssh keys that can be provided
    # via the CLI. Because their existence depends on the implementation we might run into KeyErrors.
    for argument in ["ssh_key_pair", "ssh_pub_key_path"]:
        try:
            if configuration.get_item_for_key(argument).value:
                return True
        except KeyError:
            continue
    return False


def validate_access_to_cluster(namespace, configuration):
    if is_ssh_key_specified(configuration):
        return

    cluster_password = configuration.get_item_for_key("cluster_password").value
    ssh_password_authentication = configuration.get_item_for_key("ssh_password_authentication").value
    log_cluster_password = configuration.get_item_for_key("log_cluster_password").value

    if ssh_password_authentication and (cluster_password or log_cluster_password):
        return

    # We need a way to bypass the validation, as the user might have deliver the SSH key
    # to their cluster in some other way.
    if not configuration.get_item_for_key("access_validation").value:
        log.info("Skipping cluster access validation due to '--no-access-validation' being set.")
        return

    raise ConfigLoadError(
        "Refusing to create the cluster, access to it would not be possible. "
        "Please specify either the SSH key (see --help), or "
        "'--ssh-password-authentication true' alongside either "
        "'--cluster-password <your_password>' or '--log-cluster-password' (to be able to see "
        "the randomly generated password). You can also use "
        "'--no-access-validation' to skip this check. "
        "Run with '--show-configuration' to see exactly which values have been set.")


def generate_password_if_not_set(parameter, configuration):
    password = configuration.get_item_for_key(parameter.key)
    log_cluster_password = configuration.get_item_for_key("log_cluster_password").value

    if password.value:
        return

    password.value = generate_random_cluster_password(length=24)
    text = "No custom cluster password specified. Using a random password."
    if log_cluster_password:
        text += f" Password: '{password.value}'"
    else:
        text += (
            " Use --log-cluster-password to see it."
            " If you are going to use the Bright View you have to change this password"
            " by using cm-change-passwd script (change root password of head node) or"
            " simply 'passwd root' command. If this cluster has a high-availability setup"
            " with 2 head nodes, be sure to change the password on both head nodes."
        )

    log.info(text)


def validate_password_strength(parameter, configuration):
    # a password is defined to be weak if any of the following conditions hold:
    # 1. it is 8 or fewer characters long or;
    # 2. it does not contain a digit or;
    # 3. it does not contain an upper-case letter;
    def is_weak_password(s):
        return (len(s) <= 8 or
                all(str(n) not in s for n in range(10)) or
                all(chr(65 + a) not in s for a in range(26)))

    password = configuration.get_item_for_key(parameter.key).value
    password_policy = configuration["weak_password_policy"]

    # If the password is not set we can skip the validation.
    if not password:
        return

    if is_weak_password(password):
        if password_policy == "allow":
            return

        elif password_policy == "warn":
            log.warning("The password you entered is not strong enough, "
                        "and it is being used in conjunction with enabled "
                        "SSH password authentication! Please consider "
                        "using a stronger password, or disabling SSH "
                        "password authentication. This warning can be "
                        "disabled by altering the weak password policy.")

        elif password_policy == "fail":
            raise ConfigLoadError("The password you entered is not strong enough! "
                                  "Current configuration forbids weak passwords "
                                  "in conjunction with enabled SSH password authentication. "
                                  "Use a stronger password, or disable SSH password "
                                  "authentication. A strong password must contain "
                                  "9 or more characters, out of which at least one "
                                  "is a number, and at least one is an upper-case letter.")
        else:
            raise ConfigLoadError(f"Could not determine the password policy {password_policy}. "
                                  "--weak_password_policy is not one of: allow, warn, fail.")



# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import, division

import getpass
import json
import logging
import os
import socket
import sys
import time
import uuid
from contextlib import contextmanager

import requests

import clusterondemand.utils as codutils
from clusterondemand import version
from clusterondemand.exceptions import CODException, UserReportableException
from clusterondemandconfig import config

log = logging.getLogger("cluster-on-demand")


# set by "init_tracing"
_cod_command = None
_cod_argv = None


def init_tracing(cod_command, cod_argv):
    """
    Initializes some global data which is later used by calls to send_event
    """
    global _cod_command, _cod_argv
    _cod_command = cod_command
    _cod_argv = cod_argv


class Timer(object):
    """
    Helper class for measuring elapsed time in various places of the code
    """
    def __init__(self):
        self._start_ts = time.time()
        self._stop_ts = None

    def restart(self):
        self._start_ts = time.time()
        self._stop_ts = None

    def get_elapsed(self):
        if self._stop_ts:
            until = self._stop_ts
        else:
            until = time.time()
        return round(until - self._start_ts, 2)

    def stop(self):
        if self._stop_ts:
            raise RuntimeError("Cannot stop stopped timer")
        self._stop_ts = time.time()

        return self.get_elapsed()


def get_trace_id():
    """
    Returns a UUID which is unique across each execution runtime of COD.
    Can be used to correlate logs across different system
    """
    if not hasattr(get_trace_id, "trace_id"):
        get_trace_id.trace_id = str(uuid.uuid4())

    return get_trace_id.trace_id


def log_trace_id(log_destination=log.info):
    """
    Logs the trace ID in a standard format
    """
    if "events" in config and config["events"]:
        log_destination("Trace ID: %s" % get_trace_id())


def send_event(name, **kwargs):
    """
    Sends a HTTP POST request to a specified URL.
    The request containes a json dictionary as payload, containing
    at least the "name" of the event (arbitrary string)
    """
    # This can happen upon e.g. import error, when
    # send_event is called from our exception handler
    if "events" not in config:
        log.debug("not sending event, 'events' not defined. Is global configuration defined?")
        return False

    if not config["events"]:
        return False

    if codutils.running_in_virtual_env():
        if not config["events_in_ve"]:
            log.debug("Not sending event, running in a virtual environment, use '--events-in-ve' to override.")
            return False

    data = _get_event_data(name, **kwargs)

    try:
        expected_status_code = 200
        response = requests.post(config["events_url"], data=json.dumps(data), timeout=config["events_timeout"])
        if response.status_code == expected_status_code:
            log.debug(response.text)
        else:
            raise requests.HTTPError(
                "Response status code was not %s, it was %s" % (expected_status_code, response.status_code))
    except (requests.ConnectionError, requests.ReadTimeout, requests.HTTPError) as e:
        log.debug("Failed to submit event: %s" % e)

    return True


@contextmanager
def trace_events(command_name):
    try:
        # TODO: add cloud type (openstack, AWS, Azure)
        init_tracing(cod_command=command_name, cod_argv=str(sys.argv))
        log_trace_id(log_destination=log.debug)
        version_info = version.get_version_info()
        send_event(
            "Running " + command_name,
            cod_version=version_info["version"],
            cod_branch=version_info["branch"],
            cod_git_hash=version_info["git_hash"],
            cod_git_date=version_info["git_date"],
            build_date=version_info["build_date"])
        yield
    except UserReportableException as e:
        send_event("UserReportableException in" + command_name, error=str(e))
        log_trace_id(log_destination=log.error)
        raise
    except CODException as e:
        send_event("CODException in " + command_name, error=str(e))
        log_trace_id(log_destination=log.error)
        raise
    except Exception as e:
        send_event("Unhandled Exception in " + command_name, error=str(e))
        log_trace_id(log_destination=log.error)
        raise


def _get_event_data(name, **kwargs):
    """
    Helper class of 'send_event'.
    Prepares data to be send by the event
    """
    global _cod_command, _cod_argv
    assert _cod_command
    assert _cod_argv

    assert "type" not in kwargs.keys()
    assert "username" not in kwargs.keys()
    assert "hostname" not in kwargs.keys()
    assert "trace_id" not in kwargs.keys()
    assert "ve" not in kwargs.keys()

    kwargs["name"] = name
    kwargs["type"] = "COD"
    kwargs["username"] = getpass.getuser()
    kwargs["cod_client_hostname"] = socket.getfqdn()
    kwargs["trace_id"] = get_trace_id()
    if codutils.running_in_virtual_env():
        kwargs["ve"] = True

    kwargs["cod_command"] = _cod_command
    kwargs["cod_argv"] = _cod_argv

    if "OS_USERNAME" in os.environ:
        assert "OS_USERNAME" not in kwargs.keys()
        kwargs["OS_USERNAME"] = os.environ["OS_USERNAME"]
    return kwargs



# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import re
from fnmatch import translate
from itertools import chain


def match_by_id_query(id_query, id, revision):
    """Match image data using id_query.

    >>> rev_reg = make_id_query([("id1", 1), ("id2", None)], [("pat*", 1)], ["regex.*"])

    >>> match_by_id_query(rev_reg, "id1", 1)
    True

    >>> match_by_id_query(rev_reg, "id1", 2)
    False

    >>> match_by_id_query(rev_reg, "id2", 1)
    True

    >>> match_by_id_query(rev_reg, "id2", 2)
    True

    >>> match_by_id_query(rev_reg, "pattern", 1)
    True

    """
    for rev, regex in id_query:
        if rev is not None and revision != rev:
            continue
        if id and regex.match(id) is not None:
            return True
    return False


def make_id_query(ids, patterns, regexes):
    """Return list of tuples (revision, combined_regex).

    >>> rev_reg = dict(make_id_query([("id1", 1), ("id2", None)], [("pat*", 1)], ["regex.*"]))
    >>> bool(rev_reg[1].match("id1"))
    True
    >>> bool(rev_reg[1].match("id2"))
    False
    >>> bool(rev_reg[1].match("pattern"))
    True
    >>> bool(rev_reg[1].match("regexp"))
    False

    >>> bool(rev_reg[None].match("id1"))
    False
    >>> bool(rev_reg[None].match("id2"))
    True
    >>> bool(rev_reg[None].match("pattern"))
    False
    >>> bool(rev_reg[None].match("regexp"))
    True

    """
    regex_queries = chain(
        # Creating a simple full match regex for a explicit ids ...
        (("^%s$" % (id,), rev) for id, rev in ids),
        # ... converting pattern to a regex
        ((translate(pat), rev) for pat, rev in patterns),
        # ... and using regexes as is with None revision
        ((regex, None) for regex in regexes)
    )

    result_dict = dict()
    for regex, revision in regex_queries:
        combined_regex = result_dict.get(revision, "")
        if len(combined_regex) == 0:
            combined_regex += "(?:"
        else:
            combined_regex += "|"
        combined_regex += "(?:" + regex + ")"
        result_dict[revision] = combined_regex

    return [(rev, re.compile(regex + ")")) for rev, regex in result_dict.items()]



# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import http.client as httplib
import logging
import os
import re
from urllib.parse import urlparse

import requests

from clusterondemand.clusternameprefix import get_prefix
from clusterondemand.copyfile import copy_file_set_version, fill_env_variables
from clusterondemand.exceptions import CODException
from clusterondemandconfig import config

log = logging.getLogger("cluster-on-demand")

CMD_READY_PROBE_PATH = "/root/cm/cmd_ready"
CMD_READY_PROBE = """\
#!/bin/bash
date
echo "Waiting for CMDaemon on `hostname` to update configuration files and start services..."
if [ "$(curl -m 61 -s -o /dev/null -w %{http_code} http://localhost:8080/ready?wait=60\&name=RamdiskService\&name=SysConfigGenerator\&name=SysConfigWriter\&name=dhcpd\&name=named\&name=nfs\&name=nslcd\&name=shorewall)" == "404" ]; then
    echo "The CMDaemon on `hostname` doesn't support the ready service. Will wait for 30 seconds to allow CMDaemon to set up base services."
    sleep 30
    echo "Done, will assume CMDaemon on `hostname` is ready."
    exit 0
fi
while [ "$(curl -m 61 -s -o /dev/null -w %{http_code} http://localhost:8080/ready?wait=60\&name=RamdiskService\&name=SysConfigGenerator\&name=SysConfigWriter\&name=dhcpd\&name=named\&name=nfs\&name=nslcd\&name=shorewall)" != "200" ]; do
    date
    echo "CMDaemon on `hostname` is not ready yet. Will wait some more..."
    sleep 3
done
date
echo "CMDaemon on `hostname` is ready."
"""  # noqa: E501

REMOTE_NODE_DISK_SETUP_PATH = "/root/cm/node-disk-setup.xml"
URL_REGEX = r"^http\:\/\/.*"


# When nodes are created from an image, the resulting root file-system
# will typically have the size of the image. We inject this script to
# grow the file-system to span the entire disk. This is done by finding
# any disk that has a single partition with an XFS file-system, because
# our images are created with XFS. When cluster-extension is done from
# the cod cluster, this script may end up on cloud nodes. We therefore
# skip any node-installer disks.
CATEGORY_INITIALIZE_SCRIPT_PATH = "/root/cm/category-initialize-script"
CATEGORY_INITIALIZE_SCRIPT = """\
#!/bin/bash

# This script is setup by the cluster-on-demand client. It finds disks with a
# single XFS partition (that's not the node-installer) and grows the file-
# system to span the entire disk.
disks=$(lsblk -lno name,type | grep 'disk$' | awk '{print $1}')
for disk in ${disks}; do
  echo -n "Detected disk /dev/${disk}"
  nr_of_partitions=$(lsblk -lno name,type /dev/${disk} | grep -c 'part$')
  if [ "${nr_of_partitions}" -eq "1" ]; then
    partition=$(lsblk -lno name,type /dev/${disk} | grep 'part$' | awk '{print $1}')
    # note, use 'file' instead of 'lsblk'
    # 'lsblk -lno fstype /dev/$partition' sometimes returns empty, see CM-34808
    # "SGI XFS filesystem data" string is from 'file' program source, see CM-41737
    if file --special-files --dereference /dev/$partition | grep --quiet --ignore-case "SGI XFS filesystem data"; then
      if [ "$(lsblk -lno label /dev/${partition})" != "BCMINSTALLER" ]; then
        echo ", resizing partition /dev/${partition}."
        echo "resizepart 1 -1" | parted /dev/${disk}

        echo -e "\nMounting /dev/${partition}."
        tries=5
        while [[ "${tries}" -gt "0" ]]; do
          if ! mount /dev/${partition} /mnt; then
            ((tries--))
            echo "Mounting /dev/${partition} failed, ${tries} attempts left."
            sleep 1
          else
            echo "Mounted /dev/${partition}."
            break
          fi
        done
        if [[ "${tries}" -eq "0" ]]; then
          echo "Failed to mount /dev/${partition}, aborting."
          exit 1
        fi

        echo "Resizing XFS file-system on /dev/${partition}."
        xfs_growfs /mnt

        echo "Unmounting /dev/${partition}."
        umount /mnt
      else
        echo ", partition /dev/${partition} has node-installer label, skipping resize."
      fi
    else
      echo ", partition /dev/${partition} is not XFS, skipping resize."
    fi
  else
    echo ", contains ${nr_of_partitions} partitions, skipping resize."
  fi
done
"""


def _get_node_disk_setup():
    if config["node_disk_setup_path"]:
        with open(config["node_disk_setup_path"]) as file_handle:
            log.info("Using disk setup from %s" % config["node_disk_setup_path"])
            node_disk_setup = file_handle.read()
    elif config["node_disk_setup"]:
        log.debug("Using disk setup from --node-disk-setup")
        node_disk_setup = config["node_disk_setup"]
    else:
        raise CODException("Please specify a valid node disk setup using --node-disk-setup or "
                           "--node-disk-setup-path.")
    # FIXME: need to remove newlines as somehow the file ends up
    # with '\n' characters if there is newline anywhere, which breaks the setup
    return str(node_disk_setup.replace("\n", ""))


def _get_cm_setup_conf_content():
    temp = {}

    def _ensure_structure(dictionary, module_name):
        if "modules" not in dictionary:
            dictionary["modules"] = {}
        if module_name not in dictionary["modules"]:
            dictionary["modules"][module_name] = {}

    if config["inject_cod_prefix_as_aws_provider_tag"]:
        cod_prefix = get_prefix()
        if not cod_prefix:
            raise CODException("'inject_cod_prefix_as_aws_provider_tag' is set, but the COD_PREFIX is not set")
        _ensure_structure(temp, "clusterextension")
        cluster_extension = temp["modules"]["clusterextension"]
        if "provider_tags" not in cluster_extension:
            cluster_extension["provider_tags"] = []

        cluster_extension["provider_tags"] += ["COD_PREFIX=%s" % cod_prefix]

    return temp


def validate_copy_file_parameters(bright_version, raise_on_missing=True):
    if config["copy_file_with_env"]:
        file_pairs = copy_file_set_version(config["copy_file_with_env"], bright_version)
    elif config["copy_file"]:
        file_pairs = config["copy_file"]
    else:
        return

    if config["copy_file"] or config["copy_file_with_env"]:
        compiled_url_regex = re.compile(URL_REGEX)
        for source, destination in file_pairs:
            if compiled_url_regex.match(source):
                parsed_uri = urlparse(source)
                c = httplib.HTTPConnection(parsed_uri.netloc)
                c.request("HEAD", parsed_uri.path)
                if c.getresponse().status != 200:
                    if raise_on_missing:
                        raise CODException("URL '%s' is unreachable, please make sure "
                                           "the provided URL is valid." % source)
                    else:
                        log.warn("URL '%s' is unreachable, the file will not be copied to the cluster." % source)
                else:
                    log.debug("Copying file '%s' to '%s' on the head node.", source, destination)
            else:
                if not os.path.isfile(source):
                    if raise_on_missing:
                        raise CODException(
                            "The file '%s' does not exist, please make sure the specified path is valid." % source)
                    else:
                        log.warn("The file '%s' does not exist, it will not be copied to the cluster." % source)
                elif destination.startswith("/home"):
                    log.warning("Copying file '%s' to '%s' on the head node.", source, destination)
                    log.warning("Destination file '%s' is in a home directory. Did you mean '/root'?", destination)
                else:
                    log.debug("Copying file '%s' to '%s' on the head node.", source, destination)


def _append_to_write_files(cloud_config, file_pairs, cluster_version=None, get_env=False, preserve_permissions=False):
    """Append files to the write_files in the heat stack in order to write them on the headnode.

    :param cloud_config: CloudConfig object to add the files to
    :param file_pairs: A series of either colon separated src:dst paths
    :param cluster_version: Cluster version
    :param get_env: Specify whether or not to we want to import environment variables
    :param preserve_permissions: Specify whether or not to preserve source file permissions
    """
    file_pairs = copy_file_set_version(file_pairs, cluster_version)
    compiled_url_regex = re.compile(URL_REGEX)
    for file_pair in file_pairs:
        source_file, destination_file = file_pair
        source_permissions = None

        if compiled_url_regex.match(source_file):
            r = requests.get(source_file)
            if r.status_code != 200:
                raise CODException("URL '%s' is unreachable, please make sure "
                                   "the provided URL is valid." % source_file)
            raw_content = r.text.encode()
        else:
            if not os.path.isfile(source_file):
                continue
            if preserve_permissions:
                source_permissions = "{:04o}".format(os.stat(source_file).st_mode)[-4:]
            with open(source_file, "rb") as fdata:
                raw_content = fdata.read()

        if get_env:
            raw_content = fill_env_variables(destination_file, raw_content.decode()).encode()
        cloud_config.add_file(
            destination_file,
            raw_content,
            base64=True,
            permissions=source_permissions
        )


def add_common_head_files(cloud_config, bright_version):
    cloud_config.add_file(CMD_READY_PROBE_PATH, str(CMD_READY_PROBE))
    cloud_config.add_file(REMOTE_NODE_DISK_SETUP_PATH, _get_node_disk_setup())
    cloud_config.add_file(CATEGORY_INITIALIZE_SCRIPT_PATH, CATEGORY_INITIALIZE_SCRIPT)

    cm_setup_conf = _get_cm_setup_conf_content()
    if cm_setup_conf:
        # no yaml.dump for content, as it's not clear how to force proper encoding
        # of newlines for unknown reason it works differently than e.g. for cluster_info
        cloud_config.add_file("/root/cm-setup.conf", cm_setup_conf, permissions="0600")

    if config["copy_file"]:
        _append_to_write_files(
            cloud_config,
            config["copy_file"],
            cluster_version=bright_version,
            preserve_permissions=True,
        )
        # no yaml.dump for content, as it's not clear how to force proper encoding
        # of newlines for unknown reason it works differently than e.g. for cluster_info

    if config["copy_file_with_env"]:
        _append_to_write_files(
            cloud_config,
            config["copy_file_with_env"],
            get_env=True,
            cluster_version=bright_version,
            preserve_permissions=True,
        )

    if config["license_activation_password"]:
        cloud_config.add_file(
            "/var/spool/cmd/license-activation-server.pass",
            config["license_activation_password"],
            permissions="0600", base64=True)



# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import re

from clusterondemand import const
from clusterondemand.utils import partition

log = logging.getLogger("cluster-on-demand")


def format_cluster_tags(dict_tags, unparsed_tags=None):
    if unparsed_tags is None:
        unparsed_tags = []

    list_tags, single_tags = partition(lambda x: isinstance(dict_tags[x], list), dict_tags)

    collapsed_dict_tags = ["%s=%s" % (tag, dict_tags[tag]) for tag in single_tags]
    for tag in list_tags:
        collapsed_dict_tags += ["%s#%s" % (tag, value) for value in dict_tags[tag]]

    tags = collapsed_dict_tags + unparsed_tags
    return tags


def format_packagegroups_tags(package_groups):
    return [f"{const.COD_PACKAGE_GROUP}={group}" for group in package_groups]


def parse_cluster_tags(tags):
    """Parse the cluster tags to a shape easier to handle.

    Clusters will have tags like 'COD::XX=YY', this returns a dict where dict['COD::XX'] = YY
    Tags that don't follow this key=value form, are returned in a list

    Since the package group tag can appear more than once, it's not included.
    Use get_packagegroups_from_tags(tags) to get that data.

    Tags are allowed to be repeated using the format ['COD::XX#YY', 'COD::XX#ZZ']. In that
    case it will be returned as dict['COD::XX'] = ['YY', 'ZZ']

    :param tags: list of cluster tags
    :return: Tuple (dict_tags, unparsed_tags)
    """
    if not tags:
        return ({}, [])

    single_tags_regex = re.compile("(?!" + const.COD_PACKAGE_GROUP + ")(COD::.+)=(.+)")
    list_tags_regex = re.compile("(COD::.+)#(.+)")

    single_tags = [tag for tag in tags if single_tags_regex.match(tag)]
    list_tags = [tag for tag in tags if list_tags_regex.match(tag)]

    # Add single tags
    dict_tags = {k: v for (k, v) in [single_tags_regex.match(tag).groups() for tag in single_tags]}

    # Add list tags
    for tag in list_tags:
        k, v = list_tags_regex.match(tag).groups()
        if k not in dict_tags:
            dict_tags[k] = []
        dict_tags[k].append(v)

    # Check for unparsed tags
    unparsed_tags = [tag for tag in tags if tag not in single_tags and tag not in list_tags]

    return dict_tags, unparsed_tags


def get_packagegroups_from_tags(tags):
    """Return the list of package groups from the cluster tags."""
    if not tags:
        return []
    regex = re.compile(const.COD_PACKAGE_GROUP + "=(.+)")
    return [m.group(1) for m in [regex.match(tag) for tag in tags] if m]



# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import re

import clusterondemand


def only_explicit_ids(ids, patterns, regexes, uuids, advanced):
    """Determine if query set contains only explicit image ids (names).

    A pattern is there
    >>> only_explicit_ids([("id", 1), ("id", 2)], ["pat*"], [], [], False)
    False

    A regex is there
    >>> only_explicit_ids([("id", 1), ("id", 2)], [], ["re.*"], [], False)
    False

    An id specifier does not specifies a version
    >>> only_explicit_ids([("id", 1), ("id", None)], [], [], [], False)
    False

    Which is fine for the advanced mode
    >>> only_explicit_ids([("id", 1), ("id", None)], [], [], [], True)
    True

    Now we can pick images by ids
    >>> only_explicit_ids([("id", 1), ("id", 2)], [], [], [], False)
    True

    Empty ids specified
    >>> only_explicit_ids([], [], [], [], False)
    False

    Only UUIDs specified
    >>> only_explicit_ids([], [], [], ["uuid", "uuid1"], False)
    True

    """
    if patterns or regexes:
        return False
    if not ids and not uuids:
        return False
    if not advanced and any(revision is None for _id, revision in ids):
        return False
    return True


def classify_ids(ids):
    """Parse set of passed Image IDs specifiers.

    Each of the specifiers can be either "image_id" or "image_id:revision" or "image_glance_uuid" or
    "image_id_wildcard" or "image_id_wildcard:revision" or "re:image_id_regex"

    >>> classify_ids(["image-id:1"])
    ([('image-id', 1)], [], [], [])

    >>> classify_ids(["re:image-id.*"])
    ([], [], ['image-id.*'], [])

    >>> classify_ids(["image-id.*:3"])
    ([], [('image-id.*', 3)], [], [])

    >>> classify_ids(["01234567-89ab-cdef-0123-456789abcdef"])
    ([], [], [], ['01234567-89ab-cdef-0123-456789abcdef'])

    >>> classify_ids(["ami-1234"])
    ([], [], [], ['ami-1234'])

    >>> classify_ids(["https://brightimages.blob.core.windows.net/images/bcm-cod-image-9.0-ubuntu1804-10.vhd"])
    ([], [], [], ['https://brightimages.blob.core.windows.net/images/bcm-cod-image-9.0-ubuntu1804-10.vhd'])

    >>> classify_ids(["image-id:1", "image-id.*:3", "re:image-id.*", \
        "01234567-89ab-cdef-0123-456789abcdef"])
    ([('image-id', 1)], [('image-id.*', 3)], ['image-id.*'], \
    ['01234567-89ab-cdef-0123-456789abcdef'])

    """
    plain_ids = []
    patterns = []
    regexes = []
    uuids = []
    for id in ids:
        if id.startswith("re:"):
            regexes.append(id[3:])
        elif re.match(r"[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}", id):
            uuids.append(id)
        elif re.match(r"ami-.*", id):
            uuids.append(id)
        elif re.match(r"https://.*", id):
            uuids.append(id)
        elif "*" in id or "?" in id or re.search(r".*\[.*\].*", id):
            patterns.append(parse_id_revision(id))
        else:
            plain_ids.append(parse_id_revision(id))
    return (plain_ids, patterns, regexes, uuids)


ID_REVISION_FORMAT = r"^(?P<id>[^:]+)(?::(?P<revision>[0-9]+))?$"
ID_REVISION_RE = re.compile(ID_REVISION_FORMAT)


def parse_id_revision(id_str):
    """Parse image_id:revision string.

    >>> parse_id_revision("image-id:2")
    ('image-id', 2)

    >>> parse_id_revision("image-id")
    ('image-id', None)

    >>> parse_id_revision("image-id:wrong:")  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
       ...
    CODException: image-id:wrong: has an invalid id:revision format. id should not contain colons \
    and revision has to be numeric
    """
    match = ID_REVISION_RE.match(id_str)
    if match is None:
        raise clusterondemand.exceptions.CODException(
            "%s has an invalid id:revision format. "
            "id should not contain colons and revision has to be numeric"
            % (id_str,)
        )
    id, revision_str = match.group("id"), match.group("revision")
    return id, int(revision_str) if revision_str is not None else None



# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


COD_DESCRIPTION = "cluster-on-demand"
COD_TAG = "COD::is_cluster_on_demand"
COD_TOOL_VERSION = 1
COD_TOOL_VERSION_TAG = "COD::cod_version"
COD_DISTRO_TAG = "COD::bcm_distro"
COD_VERSION_TAG = "COD::bcm_version"
COD_CMDREV_TAG = "COD::bcm_cmd_initial_revision"
COD_PACKAGE_GROUP = "COD::bcm_initial_package_group"
COD_HEAD_IMAGE_CREATED_AT = "COD::head_image_created_at"
COD_HEAD_IMAGE_ID = "COD::head_image_id"
COD_HEAD_IMAGE_NAME = "COD::head_image_name"
COD_HEAD_IMAGE_REV = "COD::head_image_revision"
COD_HEAD_IMAGE_UUID = "COD::head_image_uuid"
COD_LOCKED_UNTIL = "COD::locked_until"
COD_EDGE_ATTACHED = "COD::edge"
COD_HAS_CX = "COD::has_cluster_extension"
COD_TRACE_ID = "COD::trace_id"



# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import, print_function

import errno
import json
import logging
import socket
import ssl
import tempfile
from sys import argv as argv

import six
import six.moves.http_client
import six.moves.urllib.error
import six.moves.urllib.parse
import six.moves.urllib.request

import clusterondemand.utils
from clusterondemand.bcm_version import BcmVersion
from clusterondemand.exceptions import CODException
from clusterondemand.ssh import SSHExecutor

log = logging.getLogger("cluster-on-demand")


class CallReturnedErrorMessageError(RuntimeError):
    pass


class HTTPSClientAuthHandler(six.moves.urllib.request.HTTPSHandler):

    def __init__(self, keyfile, certfile):
        six.moves.urllib.request.HTTPSHandler.__init__(self)
        self.context = ssl.create_default_context()
        self.context.check_hostname = False
        self.context.verify_mode = ssl.CERT_NONE
        self.context.load_cert_chain(certfile, keyfile)

    def https_open(self, req):
        return self.do_open(self.get_connection, req)

    def get_connection(self, host, timeout=120):
        return six.moves.http_client.HTTPSConnection(host, context=self.context, timeout=timeout)


class CMDaemonJSONClient(object):

    def __init__(self, address, port=8081, protocol="https", ssh_identity_files=None,
                 timeout=120, cmdaemon_version=None):
        self._address = address
        self._url = "%s://%s:%d" % (protocol, address, port)
        self._ssh_identity_files = ssh_identity_files

        self._key_file = tempfile.NamedTemporaryFile()
        self._cert_file = tempfile.NamedTemporaryFile()

        self._connected = False
        self._fetched_client_cert = False
        self._opener = None
        self.timeout = timeout
        self._cmdaemon_version = BcmVersion(cmdaemon_version) if cmdaemon_version else None

    def _fetch_client_cert(self, silent=True):
        if self._fetched_client_cert:
            return

        log.debug("Fetching client certificate from %s." % self._address)
        ssh = SSHExecutor(self._address, identity_files=self._ssh_identity_files, raise_exceptions=False)
        if ssh.scp_from_remote("/root/.cm/admin.pem", self._cert_file.name) == 0 and \
                ssh.scp_from_remote("/root/.cm/admin.key", self._key_file.name) == 0 or \
                ssh.scp_from_remote("/root/.cm/cmsh/admin.pem", self._cert_file.name) == 0 and \
                ssh.scp_from_remote("/root/.cm/cmsh/admin.key", self._key_file.name) == 0:
            log.debug("Fetched admin cert, key in %s, %s." %
                      (self._cert_file.name, self._key_file.name))
            self._fetched_client_cert = True
        else:
            msg = "Failed to fetch CMDaemon client certificate from %s, please make sure your " \
                  "ssh key allows passwordless root ssh access." % self._address
            if silent:
                log.debug(msg)
            else:
                log.error(msg)
            raise CODException(msg)

    def _connect(self, silent=False):
        if self._connected:
            return True

        self._connected = True
        self._fetch_client_cert()

        log.debug("Connecting to cluster at %s." % self._address)
        self._opener = six.moves.urllib.request.build_opener(HTTPSClientAuthHandler(
            self._key_file.name, self._cert_file.name))
        try:
            self.ping()
        except six.moves.urllib.error.URLError as e:
            self._connected = False
            err_msg = "Failed to connect to CMDaemon at %s, reason: %s" % (self._address, e.reason)
            if silent:
                log.debug(err_msg)
            else:
                log.error(err_msg)
            raise CODException(err_msg, caused_by=e)

        return True

    def ready(self, timeout=None, services=None):
        """
        Check if cmdaemon is "ready".

        Older cmdaemon versions support this feature and some not. (See: CM-20007)
        If supported it will return a proper True/False status

        If not supported, True is return (i.e. We assume it's ready)

        :param timeout: [optional] Overried the timeout set in the constructor
        :param services: [optional] The list of cmdaemon services to be ready
        """
        ready = True
        detected_ready_state = False

        if timeout is None:
            timeout = self.timeout

        address = self._url + "/ready"
        if services is not None:
            # The list of services fetched by `wget -qO- http://master:8080/ready?all=yes`
            known_services = {"AutomaticFSExport", "CephConfiguratorMaster", "ConfigDumper", "GPUToolsAMD",
                              "GPUToolsDCGM", "KubernetesNodeStatusObserver", "NodeHierarchyConfiguration",
                              "ProxyService", "RamdiskService", "ResourcePoolManagerSlave", "SysConfigGenerator",
                              "SysConfigReader", "SysConfigWriter", "WLMRoleStateChange", "WlmServerCache",
                              "WlmServerInfo", "WlmSupervisorManager", "conman", "dhcpd", "mariadb", "named",
                              "nfs", "nslcd", "ntpd", "postfix", "shorewall", "shorewall6"}
            assert all(s in known_services for s in services)
            address += "?" + "&".join(f"name={s}" for s in services)

        try:
            self._connect()
            self._opener.open(address, timeout=timeout)
            detected_ready_state = True
        except six.moves.urllib.error.HTTPError as e:
            log.debug(
                "HTTP connection to {address} returned code {code}".format(
                    address=address,
                    code=e.code,
                )
            )

            # Older versions of cmdaemon do not offer the /ready endpoint, and the request for that
            # endpoint will fail with 404. In cases where we don't know which version of cmdaemon we
            # are working with, we will consider any value other than 503 to mean that the cluster
            # is ready.
            # Only when we do know the version of cmdaemon, and we do know that the version should
            # support the /ready endpoint do we consider anything other than 200 to be an indication
            # that cmdaemon is ready.
            if self._cmdaemon_version and self._cmdaemon_version >= "8.2":
                ready = False
                detected_ready_state = True
            elif e.code == 503:
                ready = False
                detected_ready_state = True
        except ssl.SSLError as e:
            if "The read operation timed out" in str(e):
                # Apparently this is the way to detect a timeout
                # See: https://github.com/python/cpython/blob/4531ec74c4a/Modules/_ssl.c#L2450
                pass
            else:
                raise CODException("Error checking CMDaemon ready state", caused_by=e)
        except socket.error as e:
            if e.errno == errno.ECONNRESET:
                pass
            else:
                raise CODException("Error checking CMDaemon ready state", caused_by=e)

        if not detected_ready_state:
            log.debug("Could not detect cmdaemon ready state. Assuming ready")

        return ready

    def ping(self):
        return self.call("cmmain", "ping", silent=True)

    def call(self, service, method, args=None, silent=False):
        self._connect(silent)

        log.debug("CMDaemon call to %s::%s(%s)" % (service, method, args))
        request = {"service": service, "call": method}
        if args:
            request["args"] = args
        try:
            data = self._opener.open(
                self._url + "/json",
                json.dumps(request).encode("utf-8"),
                timeout=self.timeout
            ).read()
            log.debug("CMDaemon call to %s::%s(%s) successfull" % (service, method, args))
            data = json.loads(data)
            if "errormessage" in data:  # Older versions of BCM, e.g. 7.0-dev, return HTTP 200 with JSON on error.
                raise CallReturnedErrorMessageError(data["errormessage"])
            else:
                return data
        except Exception:
            if silent:
                log.debug("CMDaemon call to %s::%s(%s) failed." % (service, method, args))
            else:
                log.error("CMDaemon call to %s::%s(%s) failed." % (service, method, args))
            raise

    def getComputeNodes(self):
        # TODO Implement some proper version handling
        # This hack is here temporarily for backwards compatibility.
        try:
            return self.call("cmdevice", "getComputeNodes", silent=True)
        except six.moves.urllib.error.HTTPError as e:
            if e.code == 400:
                log.debug("Couldn't find getComputeNodes. Using getSlaveNodes. Is this version < 8.2?")
                return self.call("cmdevice", "getSlaveNodes")
            raise
        except CallReturnedErrorMessageError as e:
            if "Unknown cmdevice call specified" in str(e):
                log.debug("Couldn't find getComputeNodes. Using getSlaveNodes. Is this version <= 7.0?")
                return self.call("cmdevice", "getSlaveNodes")
            raise

    def pshutdown(self, node_keys):
        # TODO Implement some proper version handling
        # This hack is here temporarily for backwards compatibility.
        RUN_PRE_HALT_OPERATIONS = True
        PRE_HALT_OPERATIONS_TIMEOUT = 300 * 1000  # from cmdaemon
        try:
            return self.call(
                "cmdevice", "pshutdown",
                [RUN_PRE_HALT_OPERATIONS, PRE_HALT_OPERATIONS_TIMEOUT, node_keys])
        except six.moves.urllib.error.HTTPError as e:
            if e.code == 400:
                log.debug("pshutdown failed. Is this version < 9.1?")
                return self.call("cmdevice", "pshutdown", [node_keys])
            raise

    def getActiveAndPassiveHeadNode(self):
        active_head_node_key, passive_head_node_key, _ = self.call("cmmain", "getActivePassiveKeys")
        results = {}
        head_nodes = self.call("cmdevice", "getHeadNodes")
        for head_node in head_nodes:
            if head_node["uniqueKey"] == active_head_node_key:
                results["active_head_node"] = head_node
            if head_node["uniqueKey"] == passive_head_node_key:
                results["passive_head_node"] = head_node
        return results

    def _runPowerOperation(self, operation, nodes):
        operation_object = {
            "baseType": "PowerOperation",
            "devices": [node["uniqueKey"] for node in nodes],
            "operation": operation
        }
        self.call("cmdevice", "powerOperation", [operation_object])

    def powerOnNodes(self, nodes):
        self._runPowerOperation("ON", nodes)

    def powerOffNodes(self, nodes):
        self._runPowerOperation("OFF", nodes)


if __name__ == "__main__":
    clusterondemand.utils.setup_logging()
    log.setLevel(logging.DEBUG)
    client = CMDaemonJSONClient(argv[1])
    if client.ping():
        print("ping ok")
    print(client.call("cmdevice", "getComputeNodes"))



#!/usr/bin/env python
# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import getpass
import logging
import logging.handlers
import os
import sys
from dataclasses import dataclass, replace
from logging import DEBUG, ERROR, INFO
from typing import Type


class CustomSyslogFormatter(logging.Formatter):
    """A formatter that can resolve %(user)s within the fmt string."""
    def __init__(self, *args, **kwargs):
        self.username = getpass.getuser()
        super(CustomSyslogFormatter, self).__init__(*args, **kwargs)

    def format(self, record):
        record.user = self.username
        return super(CustomSyslogFormatter, self).format(record)


@dataclass
class LogConfig():
    level: int
    handler: Type[logging.Handler]
    format_str: str
    short_time: bool
    formatter: Type[logging.Formatter] = logging.Formatter


CONSOLE_0_CONFIG = LogConfig(
    level=INFO,
    handler=logging.StreamHandler,
    format_str="%(asctime)s: %(levelname)8s: %(message)s",
    short_time=True
)
CONSOLE_0_WITHOUT_TIME_CONFIG = replace(CONSOLE_0_CONFIG, format_str="%(levelname)8s: %(message)s")
CONSOLE_1_CONFIG = replace(CONSOLE_0_CONFIG, level=DEBUG)
CONSOLE_2_CONFIG = replace(
    CONSOLE_1_CONFIG,
    format_str="%(asctime)s: %(name)s: %(levelname)8s: %(message)s",
    short_time=False
)
CONSOLE_3_CONFIG = replace(
    CONSOLE_2_CONFIG,
    format_str="%(asctime)s: %(name)s: %(pathname)s:%(lineno)d: %(levelname)8s: %(message)s"
)

LOGFILE_CONFIG = replace(CONSOLE_3_CONFIG, handler=logging.FileHandler)

SYSLOG_CONFIG = LogConfig(
    level=DEBUG,
    handler=logging.handlers.SysLogHandler,
    formatter=CustomSyslogFormatter,
    format_str="cluster-on-demand[%(process)d][%(user)s] %(levelname)s: %(message)s",
    short_time=False
)

CONSOLE_ERROR_ONLY_CONFIG = LogConfig(
    level=ERROR,
    handler=logging.StreamHandler,
    format_str="%(asctime)s: %(name)s: %(levelname)8s: %(message)s",
    short_time=True
)


class temporarily_logging_all_statements_to_syslog:
    def __init__(self, logger):
        self.logger = logger
        self.loggerAndOriginalLevel = []

    def __enter__(self):
        for handler in self.logger.handlers:
            self.loggerAndOriginalLevel.append((handler, handler.level))

            if isinstance(handler, logging.handlers.SysLogHandler):
                handler.setLevel(DEBUG)
            elif self.logger.level > handler.level:
                handler.setLevel(self.logger.level)

        self.loggerAndOriginalLevel.append((self.logger, self.logger.level))
        self.logger.setLevel(DEBUG)

    def __exit__(self, _type, _value, _traceback):
        for (logger, level) in self.loggerAndOriginalLevel:
            logger.setLevel(level)
        return False


def setup_logging():
    """
    Configure the root and cluster-on-demand logger and their handlers.

    COD and all of its dependencies generate log messages. We split all of these generators into two categories:
    - cluster-on-demand, and (represented here by cod_logger)
    - everything else. (the root logger).
    These two are separated by disabling the (upwards) propagation of the cod logger and are configured separately.

    There are several log handlers:
     - console, the contents of which are shown directly to the user on stderr,
     - syslog, the destination of log lines depend on the system configuration, and
     - file (if configured on the CLI)

    The desired behavior is as summarized as follows:
     - errors from cod and its dependencies are always shown on the console.
     - the console handler shows only INFO from cluster-on-demand when verbosity is 0.
       for verbosity=1 it shows all DEBUG messages from cluster-on-demand,
       for verbosity=2 it shows all DEBUG messages from the entire program and all libraries,
       for verbosity=3 and up it adds filenames and line numbers to every logged line,
    - the syslog handler sends all DEBUG messages from cluster-on-demand,
    - the file handler, if set, is similar to the console handler with verbosity=3.
    """
    config = _parse_config()

    _setup_cod_logging(config)
    _setup_root_logging(config)


def _parse_config():
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument("-v", action="count", dest="verbosity", default=0)
    parser.add_argument("--log-file", default=None)
    parser.add_argument("--log-omit-time", action="store_true")
    parsed, _ = parser.parse_known_args(sys.argv)

    if parsed.log_file is None and "COD_LOG_FILE" in os.environ:
        parsed.log_file = os.environ["COD_LOG_FILE"]
    if parsed.log_file:
        parsed.log_file = os.path.expanduser(parsed.log_file)
    if not parsed.log_omit_time and "COD_LOG_OMIT_TIME" in os.environ:
        parsed.log_omit_time = os.environ["COD_LOG_OMIT_TIME"].lower() in ["1", "true", "yes"]
    return parsed


def _setup_cod_logging(config):
    """Affects the main cod logger."""
    cod_logger = logging.getLogger("cluster-on-demand")
    cod_logger.propagate = False

    cod_logger.setLevel(DEBUG)
    cod_logger.addHandler(_console_handler(verbosity=config.verbosity, omit_time=config.log_omit_time))
    cod_logger.addHandler(_handler_for_log_config(SYSLOG_CONFIG))

    if config.log_file:
        cod_logger.addHandler(_handler_for_log_config(LOGFILE_CONFIG, filename=config.log_file))


def _setup_root_logging(config):
    """Affects logger used by libraries."""
    root_logger = logging.getLogger()

    if 2 <= config.verbosity or config.log_file:
        root_logger.setLevel(DEBUG)
    else:
        root_logger.setLevel(ERROR)

    if config.verbosity < 2:
        # From 3rd party libraries, only show errors unless -vv or -vvv
        root_logger.addHandler(_handler_for_log_config(CONSOLE_ERROR_ONLY_CONFIG))
    else:
        root_logger.addHandler(_console_handler(verbosity=config.verbosity, omit_time=False))

    if config.log_file:
        root_logger.addHandler(_handler_for_log_config(LOGFILE_CONFIG, filename=config.log_file))


def _console_handler(verbosity, omit_time):
    if 0 == verbosity:
        config = CONSOLE_0_WITHOUT_TIME_CONFIG if omit_time else CONSOLE_0_CONFIG
    elif 1 == verbosity:
        config = CONSOLE_1_CONFIG
    elif 2 == verbosity:
        config = CONSOLE_2_CONFIG
    else:
        config = CONSOLE_3_CONFIG
    return _handler_for_log_config(config)


def _handler_for_log_config(config, *handler_args, **handler_kwargs):
    formatter_kwargs = {"datefmt": "%H:%M:%S"} if config.short_time else {}
    formatter = config.formatter(config.format_str, **formatter_kwargs)

    handler = config.handler(*handler_args, **handler_kwargs)
    handler.setFormatter(formatter)
    handler.setLevel(config.level)
    return handler



# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import absolute_import, print_function

from itertools import chain

from clusterondemandconfig import ConfigNamespace, config, full_ini_config_dump, load_configuration_for_parameters

from . import configuration
from .import_utils import load_module_attribute

config_ns = ConfigNamespace("config.dump")
config_ns.import_namespace(configuration.common_ns)


COMMAND_HELP_TEXT = """
Dump the entire current configuration as an INI file. Note that this command takes into account any existing
config files, and environment variables, both of which which might impact the values in the generated config;
to avoid that, use '--no-system-config', which will generate a config from only the hardcoded defaults and
the config files specified with --config.
"""


def _load_module_parameters(module_name, commands_name):
    commands = load_module_attribute(module_name, commands_name)
    return [] if commands is None else commands.parameters()


def run_command():
    aws_parameters = _load_module_parameters("clusterondemandaws.cli", "aws_commands")
    azure_parameters = _load_module_parameters("clusterondemandazure.cli", "azure_commands")
    openstack_parameters = _load_module_parameters("clusterondemandopenstack.cli", "openstack_commands")
    vmware_parameters = _load_module_parameters("clusterondemandvmware.cli", "vmware_commands")

    config_to_dump = load_configuration_for_parameters(
        chain(aws_parameters, azure_parameters, openstack_parameters, vmware_parameters),
        system_config_files=configuration.SYSTEM_CONFIG_FILES,
        enforcing_config_files=configuration.ENFORCING_CONFIG_FILES
    )

    print(full_ini_config_dump(config_to_dump, show_secrets=config["show_secrets"]))



# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from glob import glob

from clusterondemand.images.find import CODImage, ImageSource


class LocalManifestImageSource(ImageSource):
    @classmethod
    def from_config(cls, config, ids=None):
        return LocalManifestImageSource(
            ids=ids if ids is not None else config["ids"],
            tags=config["tags"],
            version=config["version"],
            distro=config["distro"],
            package_groups=config["package_groups"],
            revision=config["revision"],
            status=config["status"],
            advanced=config["advanced"],
            image_visibility=config["image_visibility"],
            cloud_type=config["cloud_type"],
        )

    def _iter_from_source(self):
        manifest_files = find_manifest_files(self.ids)
        return (make_cod_image_from_manifest(manifest_file) for manifest_file in manifest_files)


def make_cod_image_from_manifest(manifest_file):
    """
    Wraps CODImage.from_manifest_file, implemented for readability and consistency with other imageX modules
    Constructs CODImage object from the manifest file
    :param manifest_file: path to the manifest file
    :return: CODImage object
    """
    return CODImage.from_manifest_file(manifest_file)


def find_manifest_files(ids=None, path="."):
    """
    find manifest files in the current (by default) or other directory and return their paths
    :param path: path to manifest, by default current directory. Currently not used, for future compatibility
    :param ids: list of id, revision pairs: [('ubuntu1604-9.0', 4)]
    :return: ["path/to/manifest"]
    """
    if not ids:  # If no image ids specified on the command line, detect all .manifest files
        ids = [("*", "*")]

    manifests = []
    for id_rev in ids:
        image_id, revision = id_rev
        manifests += (glob(f"{path}/*-{image_id}-{revision}*manifest"))
    return manifests



#!/usr/bin/python
# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import re
import shlex
import shutil
import subprocess
import sys
import tempfile
import textwrap
import time

import paramiko
import tenacity

from clusterondemand import localpath
from clusterondemand.codoutput.sortingutils import SSHAlias
from clusterondemand.contextmanagers import SmartLock
from clusterondemand.exceptions import CODException
from clusterondemand.utils import (
    is_valid_cluster_name,
    is_valid_ip_address,
    is_valid_positive_integer,
    is_writable_directory
)
from clusterondemandconfig import ConfigNamespace, config
from clusterondemandconfig.configuration_validation import must_be_nonnegative

clusterssh_ns = ConfigNamespace("cluster.ssh")

clusterssh_ns.add_parameter(
    "ssh_identity",
    advanced=True,
    default=None,
    help="Path to an ssh identity file key. Multiple arguments are allowed",
    help_varname="PATH_TO_FILE",
    parser=localpath.localpath,
    validation=[localpath.must_exist, localpath.must_be_readable]
)

clusterssh_ns.add_parameter(
    "ssh_password",
    advanced=True,
    default=None,
    help="SSH password to access the cluster",
    secret=True,
)

clusterssh_ns.add_parameter(
    "ssh_connect_timeout",
    advanced=True,
    default=30,
    help="Timeout for establishing an SSH connection with the cluster",
    validation=must_be_nonnegative,
)

log = logging.getLogger("cluster-on-demand")

# Suppress paramiko's logging
logging.getLogger("paramiko").setLevel(logging.WARNING)


class SSHConfigManagerException(Exception):
    """SSH configuration manager specific exceptions base class."""
    pass


class MultipleCodTypesException(Exception):
    """"""
    pass


# ssh directives (cf. man ssh_config)
HOST_ = "Host "
HOSTNAME_ = "HostName "


class SSHConfigHostDescriptor(object):
    """
    :param host: the host name, must be a valid cluster name.
    :param index: the index used to build the alias.
    :param prefix: the prefix used to build the alias.
    :param ip: the host IP, must be a valid IPv4 address.
    """
    # Notice: this ctor does not perform any validation. This is intentional, the instance owner must
    # always check the `is_valid` property to decide whether the instance should be used or discarded.
    def __init__(self, host, index, prefix, ip):
        self._host = host
        self._index = index
        self._prefix = prefix
        self._ip = ip

    @property
    def is_valid(self):
        return (is_valid_cluster_name(self.host) and
                is_valid_ip_address(self.ip) and
                is_valid_positive_integer(self.index))

    @property
    def host(self):
        return self._host

    @property
    def index(self):
        return self._index

    @property
    def ip(self):
        return self._ip

    @property
    def alias(self):
        return "{prefix}{index}".format(
            prefix=self._prefix,
            index=self._index
        )

    @staticmethod
    def from_lines(lines, prefix):
        host, index, ip = (None, ) * 3
        try:
            for line in (l.strip() for l in lines):
                if line.startswith(HOST_):
                    line = line.replace(HOST_, "").strip()

                    tmp = line.split(" ")
                    host = tmp[0]
                    index = int(tmp[1].replace(prefix, ""))

                elif line.startswith(HOSTNAME_):
                    line = line.replace(HOSTNAME_, "").strip()
                    ip = line.split(" ")[0]
        except Exception:
            pass  # silence errors

        return SSHConfigHostDescriptor(host, index, prefix, ip)

    def __str__(self):
        return textwrap.dedent("""
        Host {host} {alias}
            HostName {ip}
            User root
            UserKnownHostsFile /dev/null
            StrictHostKeyChecking=no
            CheckHostIP=no
            LogLevel=error""".format(host=self.host, alias=self.alias, ip=self.ip))


class SSHConfigManager(object):
    """
    A helper class to manage cluster-related entries in local ssh_config file.

    :param cod_type: the type of cod clusters to be managed, i.e os or vmware.
    :param ssh_config: the path of the local ssh configuration file to use.
    :param prefix: the prefix used to build the cluster ssh aliases.
    :param parse: enable parsing the contents of the ssh_config file.
    :param mode: Either "match-hosts" or "replace-hosts". "match-hosts" preserves the contents of the
                 COD section or the ssh_config file; "replace-hosts" ignores currently defined hosts.
                 User-defined contents are always preserved.
    """
    def __init__(self, cod_type, ssh_config="~/.ssh/config", prefix="", parse=True, mode="match-hosts"):

        # params
        self._ssh_config = os.path.expanduser(ssh_config)
        self._prefix = prefix
        if mode not in ["match-hosts", "replace-hosts"]:
            raise SSHConfigManagerException("Programming error")
        self._mode = mode

        # internals
        self._begin_marker = f"#### BEGIN COD-{cod_type.upper()} SECTION"
        self._end_marker = f"#### END COD-{cod_type.upper()} SECTION"
        self._cod_type = cod_type
        self._hosts = []
        self._cod_section = []
        self._usr_section = []

        self._parsed = False
        self._changed = False

        # disabling parsing of the local ssh config can be useful if the configuration is actually broken
        if parse:
            if not os.path.exists(self._ssh_config):
                # Non-existing config file is equivalent to empty config file
                log.warning("File '%s' was not found, continuing with an empty configuration.",
                            self._ssh_config)
                self._parsed = True
            else:
                try:
                    with open(self._ssh_config, "r") as config:
                        self._parse_config(config)
                        self._parsed = True

                except IOError as ioe:
                    log.warning(str(ioe))

    @staticmethod
    def lock(ssh_config="~/.ssh/config"):
        return SmartLock(os.path.expanduser(ssh_config) + ".lock")

    def _check_parsed(self):
        if not self._parsed:
            raise SSHConfigManagerException(
                "File '%s' not successfully parsed",
                self._ssh_config
            )

    @property
    def hosts(self):
        """A list of host descriptors"""
        self._check_parsed()
        return self._hosts

    @property
    def user(self):
        """Contents of the user section"""
        self._check_parsed()
        return self._usr_section

    def add_host(self, host, ip, override=False):
        """
        Add a host descriptor to current configuration. Invalid descriptors are discarded.

        :param host: the hostname (string)
        :param ip: the host IPv4 or IPv6 address (string)
        :param override: (reserved for cluster create) if another host with the same name already exists, remove the
        the corresponding entry before adding this host. If override is False, trying to add an already existing host
        will raise an exception.

        :return: the host descriptor that has been added.
        """
        self._check_parsed()

        exists = next((h for h in self._hosts if h.host == host), None)
        if exists:
            if override:
                log.warning("Replacing local ssh config entry for host '%s'", host)
                self.remove_host(host)
            else:
                raise SSHConfigManagerException("A descriptor already exists for host '%s'" % host)

        next_index = max((h.index for h in self._hosts)) + 1 if self._hosts else 1

        res = SSHConfigHostDescriptor(host, next_index, self._prefix, ip)
        self._safe_add_host_descriptor(res)

        return res

    def remove_host(self, host):
        """
        Remove a host descriptor from current configuration

        :param host: the hostname (string)
        """
        self._check_parsed()
        prev_hosts = self._hosts
        self._hosts = [h for h in prev_hosts if h.host != host]
        if self._hosts != prev_hosts:
            self._changed = True
        else:
            log.debug(
                "host descriptor for '%s' could not be found.", host)

    def write_configuration(self):
        """
        Write configuration to ssh config file.

        SSH clients give higher priority to entries towards the top of the config
        files. Therefore the COD section is put at the head of the local ssh config
        file. If no hosts are defined (i.e. no clusters), the section will be omitted.
        """
        msg = "cowardly refusing to write file '{ssh_config}'!".format(
            ssh_config=self._ssh_config
        )

        if not self._parsed:
            log.debug("%s (file was not parsed).", msg)
            return

        if not self._changed:
            log.debug("%s (no changes detected).", msg)
            return

        if not is_writable_directory(os.path.dirname(self._ssh_config)):
            log.debug("%s (directory is not writable).", msg)
            return

        assert self._parsed and self._changed
        log.debug("rewriting local ssh config file '%s' with %d COD %s.",
                  self._ssh_config, len(self._hosts),
                  "entry" if 1 == len(self._hosts) else "entries")

        try:
            # we go the extra mile of writing the new config to a temp file and only when we known
            # everything went smooth, we overwrite the pre-existing file by copying the new one onto it.

            with tempfile.NamedTemporaryFile(mode="wt") as fd:
                # Write COD section
                if self._hosts:
                    fd.write(self._begin_marker)
                    fd.write(textwrap.dedent(f"""
                    #### NOTICE: This section of the file is managed by cm-cod-{self._cod_type}. Manual changes to this section will be
                    #### overwritten next time cm-cod-{self._cod_type} cluster create, delete or list --update-ssh-config is executed."""))  # noqa

                    for descr in self._hosts:
                        # no point in keeping invalid entries here
                        if not descr.is_valid:
                            log.debug("Skipping invalid host descriptor for '%s'",
                                      descr.host or "?")
                            continue

                        fd.write(str(descr))
                        fd.write("\n")

                    fd.write(self._end_marker)
                    fd.write("\n")

                # Dump non-COD config as-is
                for line in self._usr_section:
                    fd.write(line)
                fd.flush()

                shutil.copy(fd.name, self._ssh_config)

                # issue warning if preservation of user contents semantics can not be guaranteed: we look for a
                # Host * directive, at the top of the ssh config file. If COD section exists and if this directive
                # can not be found, a warning is issued.
                if self._usr_section and self._hosts:
                    try:
                        if not re.match(r"%s\s+\*" % HOST_.lower().strip(),
                                        next((l for l in (l.strip() for l in self._usr_section if l.strip())
                                              if not l.startswith("#"))).lower()):
                            log.warning(
                                "Possibly unsafe changes were made to '{config}'. To avoid this "
                                "warning, please add a 'Host *' directive right after the end of the "
                                "COD section.".format(
                                    config=self._ssh_config
                                ))
                    except StopIteration:
                        pass

        except IOError as ioe:
            log.warning("Could not write file '%s' (%s).", self._ssh_config, str(ioe))

    def get_host_index(self, host_name):
        try:
            return next(host.index for host in self._hosts if host.host == host_name)
        except StopIteration:
            raise SSHConfigManagerException(f"{host_name} not found in ssh config")

    def get_host_alias(self, host_name):

        try:
            alias_string = next(host.alias for host in self._hosts if host.host == host_name)
            return SSHAlias(alias_string, self._prefix)
        except StopIteration:
            raise SSHConfigManagerException(f"{host_name} not found in ssh config")

    def _safe_add_host_descriptor(self, descriptor):
        """
        If a  valid descriptor is given, add it to the internal cache. Otherwise discard it.

        :param descriptor: An instance of class SSHConfigHostDescriptor
        """
        if descriptor.is_valid:
            self._hosts.append(descriptor)
            self._changed = True
            log.debug("added host descriptor for '%s'", descriptor.host)
        else:
            log.debug("discarding not well-formed descriptor for host '%s'",
                      descriptor.host or "?")

    def _parse_config(self, fd):
        """
        Parse ssh config file.

        :param fd: An open file descriptor

        The ssh config file is logically divided in two sections: the COD section contains
        all host definitions for COD clusters, along with an alias than can be used with ssh
        -like tools, and the IP to reach the head-node and a few configuration options. It is
        enclosed between the begin and end markers; the other section (i.e. anything beyond
        the COD section) is reserved to the user and is preserved as-is.
        """
        if self._mode == "match-hosts":
            self._parse_config_aux(fd)

        elif self._mode == "replace-hosts":
            self._parse_config_aux(fd)
            log.debug("regenerating contents of {config}".format(
                config=self._ssh_config
            ))
            self._hosts = []
            self._changed = True

        else:
            assert False

    def _parse_config_aux(self, fd):
        in_cod_section = False
        for line in fd:
            if line.startswith(self._begin_marker):
                if in_cod_section:
                    raise SSHConfigManagerException(
                        "Unexpected begin marker encountered")
                in_cod_section = True
                continue

            elif line.startswith(self._end_marker):
                if not in_cod_section:
                    raise SSHConfigManagerException(
                        "Unexpected end marker encountered")
                in_cod_section = False
                continue

            elif line.startswith("#### BEGIN") and not self._prefix:
                raise MultipleCodTypesException()

            if in_cod_section and line.startswith("####"):  # skip markers in COD section
                continue

            # every line goes either to the COD section or the user section
            section = self._cod_section if in_cod_section else self._usr_section
            section.append(line)

        # at EOF we're supposed to be out of the COD section
        if in_cod_section:
            raise SSHConfigManagerException(
                "Missing end marker detected")

        # This loop maintains in 'curr' a list of lines. We scan the entire file, line by line, accumulating
        # lines in 'curr'. Every time we encounter a Host declaration we want to process the accumulated lines.
        # Then, we reset the 'curr' list and continue. Once we've scanned the entire file, a few lines will
        # still be in curr. Those will be processed separately.
        curr = []
        for line in self._cod_section:
            # "Match" is not supported within the COD section
            if line.startswith(HOST_):
                if curr:
                    descriptor = SSHConfigHostDescriptor.from_lines(lines=curr, prefix=self._prefix)
                    self._safe_add_host_descriptor(descriptor)
                    curr = []
            curr.append(line)

        # final wrap-up
        if curr:
            descriptor = SSHConfigHostDescriptor.from_lines(lines=curr, prefix=self._prefix)
            self._safe_add_host_descriptor(descriptor)
            curr = []

        assert not curr, "Unexpected"
        log.debug(
            "local ssh config file '%s' parsed, COD section holds %d %s.",
            self._ssh_config, len(self._hosts),
            "entry" if 1 == len(self._hosts) else "entries")


def private_key_for_public(public_key_path):
    if not public_key_path:
        return None

    name, ext = os.path.splitext(public_key_path)
    if os.path.exists(name) and ext:
        return [name]
    else:
        return None


def validate_ssh_access(floating_ip):
    try:
        with SSHExecutor(floating_ip):
            pass
    except SSHAuthenticationError as e:
        raise CODException(textwrap.dedent(f"""
        Authentication error while connecting to root@{floating_ip}. Make sure that provided
        credentials are valid or that you have a valid passwordless SSH access to the cluster.
        Consider using --ssh-identity and/or --ssh-password parameters or using SSH Agent
        with added corresponding identity (recommended).
        If you're creating a cluster check the following options:
        --ssh-key-pair, --ssh-pub-key-path and --ssh-password-authentication.
        """), e)


class SSHExecutorError(Exception):
    pass


class SSHTimeoutError(SSHExecutorError):
    def __init__(self, host, command, stdout, stderr, timeout):
        self.host = host
        self.command = command
        self.stdout = stdout
        self.stderr = stderr
        self.timeout = timeout
        self.message = (f"Command '{self.command}' on the host {self.host} wasn't"
                        f" exited in {self.timeout} seconds")
        super().__init__(self.message)


class SSHAuthenticationError(SSHExecutorError):
    def __init__(self, host):
        self.host = host
        self.message = f"Failed to connect to the host {self.host}: authentication error"
        super().__init__(self.message)


class SSHConnectionError(SSHExecutorError):
    def __init__(self, host):
        self.host = host
        self.message = f"Failed to connect to the host {self.host}: connection error"
        super().__init__(self.message)


class SSHExecutor(object):
    """
    A utility class to interact with a cluster head node via SSH

    :param host: the host (or IP) of the head node
    :param user: the SSH user. If None (default) is used, user `root` will be used
    :param identity_files: the SSH identity files
    :param password: the SSH password
    :param raise_exceptions: If false it returns exit status wherever possible instead of
                             raising exceptions. It's needed for old code only
    :param connect_timeout: number of seconds to retry establishing a connection to the SSH server
    :param wait_between_connect_retries: number of seconds to wait between connection retries
    """

    def __init__(self, host, user=None, identity_files=None, password=None, raise_exceptions=True,
                 connect_timeout=None, wait_between_connect_retries=1):
        if identity_files is None:
            identity_files = config.get("ssh_identity")
        if identity_files is None:
            identity_files = private_key_for_public(config.get("ssh_pub_key_path"))
        if password is None:
            password = config.get("ssh_password")
        if password is None:
            password = config.get("cluster_password")
        if connect_timeout is None:
            connect_timeout = config.get("ssh_connect_timeout")

        # if some command makes use of this class but forgets to import --ssh-connect-timeout,
        # this will lead to unclear errors.
        assert type(connect_timeout) is int and connect_timeout >= 0, "timeout must be positive integer"

        self._host = host
        self._user = user or "root"
        self._identity_files = identity_files
        self._password = password
        self._raise_exceptions = raise_exceptions
        self._keep_connection = False
        self._ssh_client = None
        self._last_error = None
        self._connect_timeout = connect_timeout
        self._wait_between_connect_retries = wait_between_connect_retries

    def __enter__(self):
        self._connect()
        self._keep_connection = True
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._keep_connection = False
        self._close()

    @property
    def last_error(self):
        """ Get last error raised by ssh client """
        return self._last_error

    def get_ssh_client(self):
        """ Get underlying SSH client. Must be called only inside of the 'with' statement """
        assert self._keep_connection
        return self._ssh_client

    def _close(self):
        if self._ssh_client:
            self._ssh_client.close()
            self._ssh_client = None

    def _try_connect_using_ssh_client(self):
        # This is a small optimization to cache passphrases in SSH Agent and avoid
        # paramiko's bug with asking a passpharse for each private key
        args = [
            "ssh", "-q",
            "-o", "PasswordAuthentication=no",
            "-o", "StrictHostKeyChecking=no",
            "-o", "UserKnownHostsFile=/dev/null",
            "-o", "ConnectTimeout=1",
            "-o", "BatchMode=yes",  # Avoid stucks because of waiting for keyboard input
            f"{self._user}@{self._host}",
            "exit"
        ]
        try:
            subprocess.check_call(args)
        except subprocess.CalledProcessError as e:
            # This is probably not an error, maybe password auth will work
            # Caller will handle connection/auth errors itself
            log.debug("Authentication using SSH client failed: %s", e)
        else:
            log.debug("Authentication using SSH client was successful")

    def _connect(self):
        if self._ssh_client:
            return

        self._try_connect_using_ssh_client()

        assert self._ssh_client is None
        self._ssh_client = paramiko.SSHClient()
        self._ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        connected = False

        @tenacity.retry(
            stop=tenacity.stop_after_delay(self._connect_timeout),
            wait=tenacity.wait_fixed(self._wait_between_connect_retries),
            before_sleep=tenacity.before_sleep_log(log, logging.DEBUG),
            reraise=True,
        )
        def connect_wait():
            self._ssh_client.connect(
                self._host,
                username=self._user,
                key_filename=self._identity_files,
                password=self._password,
                timeout=self._connect_timeout,
            )

        try:
            log.debug(f"Creating an SSH connection to {self._user}@{self._host} (timeout {self._connect_timeout}s)")
            connect_wait()
            connected = True
            log.debug(f"SSH connection to {self._user}@{self._host} established")
        except paramiko.AuthenticationException as e:
            self._last_error = e
            raise SSHAuthenticationError(self._host) from e
        except Exception as e:
            self._last_error = e
            raise SSHConnectionError(self._host) from e
        finally:
            if not connected:
                self._ssh_client = None

    def run(self, command, timeout=1800, max_stdout_size=100 * 2**20, max_stderr_size=100 * 2**20,
            capture_out=False, capture_err=False):
        if isinstance(command, list):
            command = " ".join(shlex.quote(arg) for arg in command)

        stdin, stdout, stderr, channel = None, None, None, None
        out, err = [], []
        try:
            self._connect()
            stdin, stdout, stderr = self._ssh_client.exec_command(command)
            channel = stdin.channel  # Channel is the same for stdin/stdout/stderr

            # Close write as we don't have any input
            channel.shutdown_write()

            out_size, err_size = 0, 0
            timed_out = False
            end_time = time.monotonic() + timeout
            while time.monotonic() < end_time:
                if channel.recv_ready():
                    if capture_out and out_size < max_stdout_size:
                        chunk = channel.recv(max_stdout_size - out_size)
                        out_size += len(chunk)
                        out.append(chunk)
                    else:
                        chunk = channel.recv(2**20)
                        if not capture_out:
                            sys.stdout.write(chunk.decode("utf-8", errors="replace"))
                        else:
                            pass  # Consume but do nothing to avoid blocking the command
                if channel.recv_stderr_ready():
                    if capture_err and err_size < max_stderr_size:
                        chunk = channel.recv_stderr(max_stderr_size - err_size)
                        err_size += len(chunk)
                        err.append(chunk)
                    else:
                        chunk = channel.recv_stderr(2**20)
                        if not capture_err:
                            sys.stderr.write(chunk.decode("utf-8", errors="replace"))
                        else:
                            pass  # Consume but do nothing to avoid blocking the command
                if channel.closed or channel.exit_status_ready():
                    break
                time.sleep(0.01)
            else:
                timed_out = True

            out = b"".join(out)
            err = b"".join(err)
            status = channel.recv_exit_status() if not timed_out else 255

            if status != 0:
                if status not in range(256):
                    log.debug("recv_exit_status returned unexpected status value: %d."
                              " It will be replaced by 255", status)
                    status = 255
                log.debug("SSH command '%s' returned status %d%s%s", command, status,
                          (f", stdout: {out[:4096]}" if capture_out else ""),
                          (f", stderr: {err[:4096]}" if capture_err else ""))

            if timed_out:
                raise SSHTimeoutError(self._host, command, out, err, timeout)

            return status, out, err
        except SSHTimeoutError as e:
            log.debug("SSH command timed out: %s", e)
            if not self._raise_exceptions:
                return 255, e.stdout, e.stderr
            raise
        except Exception as e:
            log.debug("Failed to execute SSH command '%s': %s", command, e)
            self._last_error = e
            if not self._raise_exceptions:
                return 255, b"", b""
            raise
        finally:
            if stdin is not None:
                stdin.channel.close()
                stdin.close()
                stdout.close()
                stderr.close()
            if not self._keep_connection:
                self._close()

    def call(self, command, **run_kwargs):
        return self.run(command, **run_kwargs)[0]

    def check_call(self, command, **run_kwargs):
        status, _, _ = self.run(command, **run_kwargs)
        if status != 0:
            raise subprocess.CalledProcessError(status, command)

    def check_output(self, command, **run_kwargs):
        status, stdout, _ = self.run(command, capture_out=True, **run_kwargs)
        if status != 0:
            raise subprocess.CalledProcessError(status, command, output=stdout)
        return stdout

    def scp_to_remote(self, src, dst, preserve_perms=False):
        try:
            self._connect()
            with self._ssh_client.open_sftp() as sftp:
                sftp.put(src, dst)
                if preserve_perms:
                    stat = os.stat(src)
                    sftp.chmod(dst, stat.st_mode)
                    sftp.utime(dst, (stat.st_atime, stat.st_mtime))
            return 0
        except SSHExecutorError as e:
            log.debug("SCP to %s failed: %s", dst, e)
            self._last_error = e
            if not self._raise_exceptions:
                return 255
            raise
        except Exception as e:
            log.debug("SCP to %s failed: %s", dst, e)
            self._last_error = e
            if not self._raise_exceptions:
                return 1
            raise
        finally:
            if not self._keep_connection:
                self._close()

    def scp_from_remote(self, src, dst):
        try:
            self._connect()
            with self._ssh_client.open_sftp() as sftp:
                sftp.get(src, dst)
            return 0
        except SSHExecutorError as e:
            log.debug("SCP from %s failed: %s", src, e)
            self._last_error = e
            if not self._raise_exceptions:
                return 255
            raise
        except Exception as e:
            log.debug("SCP from %s failed: %s", src, e)
            self._last_error = e
            if not self._raise_exceptions:
                return 1
            raise
        finally:
            if not self._keep_connection:
                self._close()

    def scp_and_call(self, script_path):
        if self.scp_to_remote(script_path, "/tmp/caas-script"):
            log.error("Failed to scp %s to %s." % (script_path, self.host))
            return False

        if self.call("chmod 700 /tmp/caas-script && /tmp/caas-script && rm /tmp/caas-script"):
            log.error("Failed to run %s on %s." % (script_path, self.host))
            return False

        return True

    def scp_and_call_on_chroot(self, script_path, root):
        temp_script = "tmp/caas-script"
        dest_path = os.path.join(root, temp_script)

        if self.scp_to_remote(script_path, dest_path):
            log.error("Failed to scp %s to %s." % (script_path, self.host))
            return False

        cmd = "chroot {root} /bin/bash -c 'chmod 700 {script} && {script} && rm {script}'".format(
            root=root, script=os.path.join("/", temp_script))
        if self.call(cmd):
            log.error("Failed to run %s on %s." % (script_path, self.host))
            return False

        return True



# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import, division

from functools import reduce

import six

from clusterondemand.codoutput.sortingutils import SortableData
from clusterondemand.utils import get_time_ago

ALL_COLUMNS = [
    ("id_revision", "ImageID:Revision"),
    ("name", "Image name"),
    ("type", "Image type"),
    ("head_size", "Head(GB)"),
    ("node_size", "Node(GB)"),
    ("size", "Size(GB)"),
    ("image_size", "Image(GB)"),
    ("distro", "Distro"),
    ("cmd_revision", "CMD Rev."),
    ("bcm_version", "BCM Version"),
    ("created_at", "Created"),
    ("uploaded_at", "Uploaded"),
    ("package_groups", "Packages"),
    ("is_public", "Public"),
    ("uuid", "UUID"),
    ("cloud_type", "Cloud Type"),
]

ADVANCED_COLUMNS = [
    "name", "type", "size", "distro", "cmd_revision", "bcm_version", "created_at", "uploaded_at",
    "package_groups", "is_public", "uuid", "cloud_type",
]

BASIC_COLUMNS = [
    "id_revision", "head_size", "node_size", "distro", "cmd_revision", "bcm_version", "created_at",
    "package_groups", "is_public"
]

USER_READABLE_IMAGE_TYPES = {
    "node": "Compute node",
    "headnode": "Head node",
    "node-installer": "Node installer",
    "edge-iso": "Edge ISO",
}


def user_readable_image_type(image_type):
    return USER_READABLE_IMAGE_TYPES.get(image_type) or ""


def make_images_table(images, sortby=None, columns=None, output_format=None):
    sortby = sortby or []
    columns = columns or [column[0] for column in ALL_COLUMNS]
    rows = [
        [getattr(image, column[0], "N/A") for column in ALL_COLUMNS]
        for image in images
    ]

    table = SortableData(all_headers=ALL_COLUMNS, requested_headers=columns, rows=rows)
    table.sort(*sortby)
    return table.output(output_format=output_format)


def make_cod_images_table(cod_images, sortby=None, columns=None, advanced=False, output_format="table"):
    if not columns:
        columns = ADVANCED_COLUMNS if advanced else BASIC_COLUMNS

    return make_images_table(
        [CODImageAdapter(image, advanced) for image in cod_images],
        sortby=sortby,
        columns=columns,
        output_format=output_format
    )


class CODImageAdapter(object):
    """Adapter that maps every key in ALL_COLUMNS to a value of a COD Image."""

    def __init__(self, image, advanced=False):
        self.name = image.name
        if image.bcm_optional_info:
            self.id_revision = "{0.id}:{0.revision} ({0.bcm_optional_info})".format(image)
        else:
            self.id_revision = "{0.id}:{0.revision}".format(image)
        self.revision = image.revision

        if 0 == len(image.node_images):
            node_images_size = None
        else:
            node_images_size = reduce(lambda acc, cur: acc + cur, [i.size for i in image.node_images], 0)
        self.size = GigabyteSize(image.size + (node_images_size or 0))
        self.head_size = GigabyteSize(image.size)
        self.node_size = GigabyteSize(node_images_size)
        self.image_size = GigabyteSize(image.size)  # used when we don't want to use "head_size" column name
        # for "flattened" images that may not be the headnode

        self.distro = image.distro
        self.cmd_revision = image.cmd_revision
        self.bcm_version = image.version
        self.created_at = RelativeDateTime(image.created_at)
        self.uploaded_at = RelativeDateTime(image.uploaded_at)
        self.package_groups = ",".join(image.package_groups)
        self.is_public = image.image_visibility == "public"
        self.uuid = image.uuid
        self.cloud_type = image.cloud_type

        image_types = [image.type]
        if not advanced:
            # In the non-advanced we show the image sets, so we put all the types here
            image_types += [node_image.type for node_image in image.node_images]
        self.type = ", ".join(user_readable_image_type(t) for t in image_types)


class GigabyteSize(object):
    """Utility class that allows both sorting by value while still giving the value a custom string representation."""

    N_BYTES_PER_GB = 1024 ** 3

    def __init__(self, bytes):
        self.bytes = bytes

    if six.PY2:
        def __cmp__(self, other):
            return cmp(self.bytes, other.bytes)  # noqa: F821
    elif six.PY3:
        def __lt__(self, other):
            return self.bytes < other.bytes

    def __str__(self):
        if self.bytes is None:
            return "N/A"
        else:
            return str(round(self.bytes / GigabyteSize.N_BYTES_PER_GB, 2))


class RelativeDateTime(object):
    """Utility class that allows both sorting by value while still giving the value a custom string representation."""

    def __init__(self, time):
        self.time = time

    if six.PY2:
        def __cmp__(self, other):
            return cmp(self.time, other.time)  # noqa: F821
    elif six.PY3:
        def __lt__(self, other):
            return self.time < other.time

    def __str__(self):
        return get_time_ago(self.time) + " ago" if self.time else "N/A"



from six.moves import map

# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class BcmVersion(object):
    """A utility class for comparing Bright Cluster Manager versions.

    Wraps a version number string and handles comparison logic. The string cannot contain a revision
    suffix, yet.

    Hard-coded assumption: At Bright, once we start working on BCM with major version X, then we
    will never release another minor version of a major X-1. That is: there will never be a release
    of something like 8.4 if 9.0 has been released (even internally).

    Also, it doesn't consider the tag after the dash (i.e: 8.1-2). This can be implemented in future if
    we see it as necessary
    """

    @classmethod
    def _wrap_with_instance(self, obj):
        return obj if isinstance(obj, self) else self(obj)

    def __init__(self, string):

        self.string = string
        self.is_trunk, self.major, self.minor = None, None, None

        if "trunk" == string:
            self.is_trunk = True
        else:
            major_dot_minor = string.split("-")[0]
            self.major, self.minor = map(int, major_dot_minor.split("."))

    def __str__(self):
        return self.string

    def __eq__(self, other):
        other = self.__class__._wrap_with_instance(other)
        return (
            self.is_trunk == other.is_trunk and
            self.major == other.major and
            self.minor == other.minor
        )

    def __ne__(self, other):
        return not self == other

    def __gt__(self, other):
        other = self.__class__._wrap_with_instance(other)
        if self.is_trunk:
            return not other.is_trunk
        elif other.is_trunk:
            return False
        else:
            return self.major > other.major or \
                (self.major == other.major and self.minor > other.minor)

    def __ge__(self, other):
        return self == other or self > other

    def __lt__(self, other):
        return not self >= other

    def __le__(self, other):
        return not self > other



# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def match_by_package_groups(package_groups, image_package_groups):
    """Check if image_package_groups fits the requirement.

    If there is "any" in package_groups than we accepting any image
    >>> match_by_package_groups(["any", "something"], ["other", "yada"])
    True

    If there is "none" in package_groups than we accept only the image with empty
    list of package groups
    >>> match_by_package_groups(["none", "something"], ["other"])
    False
    >>> match_by_package_groups(["none", "something"], [])
    True

    In other cases we are looking for images with all the passed package groups
    installed
    >>> match_by_package_groups(["gr1", "gr2"], ["gr1", "other"])
    False
    >>> match_by_package_groups(["gr1", "gr2"], ["other"])
    False
    >>> match_by_package_groups(["gr1", "gr2"], ["gr2", "gr1", "other"])
    True

    """
    if "none" in package_groups:
        return len(image_package_groups) == 0
    if "any" in package_groups:
        return True
    for package_group in package_groups:
        if package_group not in image_package_groups:
            return False
    return True



# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from abc import ABC, abstractmethod
from base64 import b64encode


class SignalHandler(ABC):
    def __init__(self, status_log_prefix: str):
        self.log_prefix = status_log_prefix
        super().__init__()

    @abstractmethod
    def get_init_commands(self):
        pass

    @abstractmethod
    def get_files(self):
        pass

    @abstractmethod
    def get_status_log_commands(self, status: str):
        pass

    @abstractmethod
    def get_config_complete_commands(self):
        pass


class CloudConfig(dict):
    def __init__(self, signal_handler: SignalHandler):
        self.signal_handler = signal_handler
        self["write_files"] = []
        self["bootcmd"] = []
        self["runcmd"] = signal_handler.get_init_commands()
        self["disable_ec2_metadata"] = False
        for file in signal_handler.get_files():
            self.add_file(file["path"], file["content"])

    def add_status_commands(self, status: str):
        self.add_commands(self.signal_handler.get_status_log_commands(status))

    def add_config_complete_commands(self):
        self.add_commands(self.signal_handler.get_config_complete_commands())

    def enable_metadata(self, enable_metadata):
        self["disable_ec2_metadata"] = not enable_metadata

    def add_file(self, path, content, base64=False, permissions=None):
        write_file = {
            "path": path
        }

        if base64:
            write_file["encoding"] = "base64"
            write_file["content"] = b64encode(content)
        else:
            write_file["content"] = content

        if permissions is not None:
            write_file["permissions"] = permissions

        self["write_files"].append(write_file)

    def add_boot_command(self, command):
        self["bootcmd"].append(command)

    def add_boot_commands(self, commands):
        self["bootcmd"] += commands

    def add_command(self, command):
        self["runcmd"].append(command)

    def add_commands(self, commands):
        self["runcmd"] += commands

    def to_dict(self):
        result = copy.deepcopy(self)
        return dict(result)



# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Blobs of text common to multiple packages, or simply too long."""
from __future__ import absolute_import

import textwrap

HR = "-" * 69

CONFIG_GENERATE_DESC = textwrap.dedent("""
This command generates a config files to stdout.
Redirect the stdout to a file to save it as a config file. Run the tool with Use '-c <path>' to
use it later. Save to ~/cm-cluster-on-demand.conf to always load it automatically.
""")

CONFIG_GENERATE_MINIMAL = textwrap.dedent("""
Only include the minimal subset of values in the config files. Value included are defaults which
typically have to be configure. The omitted values will use defaults
""")

CONFIG_GENERATE_RUNTIME = textwrap.dedent("""
Embed the runtime config values (determined from other config files) in the
config, instead of the basic defaults. Useful for enforcing the current runtime values event if
the underlying config values change.
""")


CONFIG_SHOW = textwrap.dedent("""
This command display a preview of the current runtime configuration. That is, after loading
any present config files, environment variables, command line arguments, and overlaying all
of those on each other.
""")


OPENSTACK_CLUSTER_CREATE_DESC = textwrap.dedent("""
This command creates a new cluster in OpenStack. It will create the head node, compute nodes,
and the networking required for the cluster to work.

By default the name of the cluster is autogenerated from the selected/implied version
and distribution, and optionally appended with a label.

If you want to force a full name for a cluster, use --name
""")



# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from clusterondemandconfig import ConfigNamespace, config, generate_tabcompletion_for_command_context

from . import configuration, exceptions, import_utils

TOOL_MAPPING = {
    "cm-cod": ("clusterondemand.cli", "cod_commands"),
    "cm-cod-os": ("clusterondemandopenstack.cli", "openstack_commands"),
    "cod": ("clusterondemandopenstack.cli", "openstack_commands"),
    "cm-cod-aws": ("clusterondemandaws.cli", "aws_commands"),
    "cm-cod-azure": ("clusterondemandazure.cli", "azure_commands"),
    "cm-cod-vmware": ("clusterondemandvmware.cli", "vmware_commands")
}

COMMAND_HELP_TEXT = """
Generate a bash tab-completion script that can be loaded into your current bash session or stored as
a file in the sytem bash_completion.d folder. When this script is active, you can press <tab> to
have bash complete groups, commands, flags as well as certain possible values for certain parameters.

If bash cannot resolve the prefix to a single value, then all possible completions will be shown.

Examples:
  cm-cod-os cl<tab> -> cm-cod-os cluster,
  cm-cod-os cluster cr<tab> -> cm-cod-os cluster create,
  cm-cod-os cluster create --he<tab> -> cm-cod-os cluster create --help,
"""

config_ns = ConfigNamespace("tabcompletion.generate")
config_ns.import_namespace(configuration.common_ns)
config_ns.add_positional_parameter(
    "tool",
    # TODO: CM-25743, add choices attribute to positional parameter
    help="The tool for which the tab completion script is generated",
    require_value=True,
)


def run_command():
    if config["tool"] not in TOOL_MAPPING.keys():
        raise exceptions.CODException(f"{config['tool']} must be one of {', '.join(TOOL_MAPPING.keys())}")

    module, commands = TOOL_MAPPING[config["tool"]]
    command_context = import_utils.load_module_attribute(module, commands)

    if not command_context:
        raise exceptions.CODException(f"Could not import {module}")

    generate_tabcompletion_for_command_context(config["tool"], command_context)



# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from logging import getLogger

from .exceptions import ValidationException
from .utils import is_valid_cluster_name

UUID_REGEX = "^[0-9a-fA-F]{8}(-[0-9a-f]{4}){3}-[0-9a-f]{12}$"
PRODUCT_KEY_REGEX = "^[0-9]{6}(-[0-9]{6}){4}$"
_IMAGE_NAME_REGEX = r"[^-]+"
_BCM_VERSION_REGEX = r"(?:trunk|\d+\.\d+)(?:-[a-z-]+[a-z])?"
_IMAGE_REVISION_REGEX = r"\d+"
VALID_IMAGE_REGEX = r"^(?!bcmn|bcmh)%s-%s:%s$" % (_IMAGE_NAME_REGEX, _BCM_VERSION_REGEX, _IMAGE_REVISION_REGEX)
VALID_HEADNODE_IMAGE_REGEX = r"^bcmh-%s-%s-%s$" % (_IMAGE_NAME_REGEX, _BCM_VERSION_REGEX, _IMAGE_REVISION_REGEX)
VALID_NODE_IMAGE_REGEX = r"^bcmn-%s-%s-%s$" % (_IMAGE_NAME_REGEX, _BCM_VERSION_REGEX, _IMAGE_REVISION_REGEX)

log = getLogger("cluster-on-demand")


# TODO Get rid of this class and put in functions?
# There is no point in having a class with only static methods
class ParamValidator(object):

    @staticmethod
    def validate_cluster_name(cluster_name):
        if not is_valid_cluster_name(cluster_name):
            raise ValidationException(
                f"The cluster name '{cluster_name}' does not match the proper format. "
                f"The cluster name cannot exceed 64 characters in length and "
                f"should be a valid hostname as specified by RFC952 and RFC1123."
            )

    @staticmethod
    def validate_uuid_format(uuid_string, error_message=""):
        if not error_message:
            error_message = f"{uuid_string} is not a valid UUID"

        if not bool(re.match(UUID_REGEX, uuid_string)):
            raise ValidationException(error_message)

    @staticmethod
    def validate_license_product_key(parameter, configuration):
        product_key_string = configuration[parameter.key]

        if not re.match(PRODUCT_KEY_REGEX, product_key_string):
            raise ValidationException("License product key does not match the proper format")


def validate_image_value(parameter, config):
    value = config[parameter.key]
    if value and not re.match(VALID_IMAGE_REGEX, value):
        log.warn(f"'{value}' does not seem to be a valid image. "
                 "Expected format <distro>-<bcm version>:<n>, e.g. 'ubuntu1604-trunk:13'")


def validate_headnode_image_value(parameter, config):
    value = config[parameter.key]
    if value and not re.match(VALID_HEADNODE_IMAGE_REGEX, value):
        log.warn(f"'{value}' does not seem to be a valid head node image. "
                 "Expected format bcmh-<distro>-<bcm version>-<n>, e.g. 'bcmh-ubuntu1604-trunk-13'")


def validate_node_image_value(parameter, config):
    value = config[parameter.key]
    if value and not re.match(VALID_NODE_IMAGE_REGEX, value):
        log.warn(f"'{value}' does not seem to be a valid node image. "
                 "Expected format bcmn-<distro>-<bcm version>-<n>, e.g. 'bcmn-ubuntu1604-trunk-13'")



# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import os

from .exceptions import UserReportableException


def localpath(string):
    """A parser for paths that exist on the local machine."""
    return os.path.expanduser(string)


def must_exist(parameter, configuration):
    if configuration[parameter.key] and not os.path.isfile(configuration[parameter.key]):
        raise UserReportableException(
            "{name}={value} does not exist".format(name=parameter.key, value=configuration[parameter.key])
        )


def must_be_readable(parameter, configuration):
    if configuration[parameter.key] and not os.access(configuration[parameter.key], os.R_OK):
        raise UserReportableException(
            "{name}={value} is not readable".format(name=parameter.key, value=configuration[parameter.key])
        )



# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import fnmatch
import logging
import re

from clusterondemand.clusternameprefix import clusterprefix_ns, ensure_cod_prefix
from clusterondemand.exceptions import CODException
from clusterondemandconfig import ConfigNamespace, config

from . import const

log = logging.getLogger("cluster-on-demand")


class ClusterFinder:
    """
    Class to abstract all finding functions coming from the backend. Each provider has to implement this
    """
    def __init__(self, min_cod_version=None, ignore_newer=None, all_stacks=None):
        self.all_stacks = all_stacks if all_stacks is not None else config["all_stacks"]
        self.min_cod_version = min_cod_version if min_cod_version is not None else 1
        self.ignore_newer = ignore_newer if ignore_newer is not None else config["ignore_newer"]

    def find_by_names(self, names):
        raise NotImplementedError()

    def find_by_uuids(self, uuids):
        raise NotImplementedError()

    def find_all(self):
        raise NotImplementedError()

    def find_by_tags(self, tags=None, tags_any=None, not_tags=None, not_tags_any=None):
        raise NotImplementedError()

    def get_cluster_name(self, cluster):
        raise NotImplementedError()

    def get_cluster_id(self, cluster):
        raise NotImplementedError()

    def get_cluster_tags(self, cluster):
        raise NotImplementedError()

    def filter_by_cod_version(self, clusters):
        return (
            cluster for cluster in clusters
            if self.check_cod_tool_version(cluster)
        )

    def check_cod_tool_version(self, cluster, current_cod_version=const.COD_TOOL_VERSION):
        tags = self.get_cluster_tags(cluster)
        if self.all_stacks and const.COD_TAG not in tags:
            return True

        version_re = re.compile(const.COD_TOOL_VERSION_TAG + "=([0-9]+)$")
        for tag in tags:
            m = version_re.match(tag)
            if m:
                if int(m.group(1)) < self.min_cod_version:
                    raise CODException(
                        "Cluster %s was created with an older version of the "
                        "cluster-on-demand tools. Can't perform the requested operation "
                        "on this cluster." % self.get_cluster_name(cluster)
                    )
                elif int(m.group(1)) > current_cod_version:
                    if self.ignore_newer:
                        return False
                    else:
                        raise CODException(
                            "Cluster %s was created with a newer version of the "
                            "cluster-on-demand tools. Please update your tools, or pass "
                            "--ignore-newer to ignore this cluster." % self.get_cluster_name(cluster)
                        )
                else:
                    return True

        if not self.ignore_newer:
            raise CODException(
                "Unable to determine cluster-on-demand tool version used "
                "to create cluster %s." % self.get_cluster_name(cluster)
            )
        return False

    def find_clusters_with_config(self):
        return self.find_clusters(
            filters=config["filters"],
            tags=config["tags"],
            not_tags_any=config["not_tags_any"],
            not_tags=config["not_tags"],
            version=config["version"],
            distro=config["distro"],
        )

    def find_clusters(self, filters=None, tags=None, tags_any=None, not_tags=None, not_tags_any=None,
                      version=None, distro=None):
        filters = [] if filters is None else filters
        tags = [] if tags is None else tags
        tags_any = [] if tags_any is None else tags_any
        not_tags = [] if not_tags is None else not_tags
        not_tags_any = [] if not_tags_any is None else not_tags_any

        cp_filters = filters[:]
        cp_tags = tags[:]

        if version:
            cp_tags = _replace_or_add_tag(cp_tags, const.COD_VERSION_TAG, version)
        if distro:
            cp_tags = _replace_or_add_tag(cp_tags, const.COD_DISTRO_TAG, distro)

        if not (cp_filters or cp_tags or tags_any or not_tags or not_tags_any):
            cp_filters = ["*"]

        stacks = []
        missing_names = set()

        if cp_tags or tags_any or not_tags or not_tags_any:
            stacks = self.find_by_tags(cp_tags, tags_any, not_tags, not_tags_any)
        if cp_filters:
            (stacks, missing_names) = _get_clusters_by_names(self, cp_filters, self.all_stacks, stacks)

        if len(missing_names) != 0:
            log.debug("Not found: %s" % ", ".join(sorted(missing_names)))

        return (sorted(stacks, key=lambda x: self.get_cluster_name(x)), sorted(missing_names))


def is_filtering_enabled():

    keys = ["filters", "tags", "not_tags_any", "not_tags", "version", "distro", "ignore_newer"]
    for v in keys:
        if config[v]:
            return True

    return False


def _replace_or_add_tag(tags, tag_name, tag_val):
    tag_name_eq = tag_name + "="
    cleaned_tags = [x for x in tags if not x.startswith(tag_name_eq)]
    cleaned_tags.append(tag_name_eq + tag_val)
    return cleaned_tags


def _classify_names(names):
    plain_names = set()
    patterns = []
    regexes = []
    uuids = set()
    for name in names:
        if name.startswith("re:"):
            regexes.append(ensure_cod_prefix(name[3:]))
        elif "*" in name or "?" in name or re.search(r".*\[.*\].*", name):
            patterns.append(ensure_cod_prefix(name))
        elif re.match(r"[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}", name):
            uuids.add(name)
        else:
            plain_names.add(ensure_cod_prefix(name))
    return (plain_names, patterns, regexes, uuids)


def _get_clusters_by_names(finder, names, all_stacks=False, stacks=None):
    (plain_names, patterns, regexes, uuids) = _classify_names(names)
    found_stacks = []
    if len(patterns) == 0 and len(regexes) == 0:
        if stacks:
            found_stacks = [s for s in stacks if finder.get_cluster_name(s) in plain_names or s.id in uuids]
        else:
            if len(plain_names) > 0:
                found_stacks.extend(finder.find_by_names(plain_names))
            if len(uuids) > 0:
                found_stacks.extend(finder.find_by_uuids(uuids))
    else:
        valid_regexes = []
        for regex in regexes:
            try:
                re.compile(regex)
            except re.error as e:
                raise CODException("Pattern '%s' is not a valid regex" % regex, caused_by=e)
            else:
                valid_regexes.append(regex)

        valid_regexes.extend([fnmatch.translate(p) for p in patterns])
        valid_regexes.extend([n + "$" for n in plain_names])

        combined_regex = re.compile("(?:%s)" % "|".join(valid_regexes))

        if all_stacks:
            kwargs = {}
        else:
            kwargs = {"tags": [const.COD_TAG]}

        if not stacks:
            stacks = finder.find_by_tags(**kwargs)

        def stack_ok(stack):
            return combined_regex.match(finder.get_cluster_name(stack)) or finder.get_cluster_id(stack) in uuids

        found_stacks = [s for s in stacks if stack_ok(s)]

    found_stacks = [s for s in found_stacks if finder.check_cod_tool_version(s)]
    found_names = set(finder.get_cluster_name(s) for s in found_stacks)
    found_uuids = set(finder.get_cluster_id(s) for s in found_stacks)
    missing = (plain_names - found_names) | (uuids - found_uuids)
    return (found_stacks, missing)


findclusters_ns = ConfigNamespace("common.cluster.find", help_section="cluster filter parameters")
findclusters_ns.import_namespace(clusterprefix_ns)
findclusters_ns.add_repeating_positional_parameter(
    "filters",
    default=None,
    help=("List of cluster names or UUID's. Regular "
          "expressions can be used with a 're:' prefix. Shell like wildcards are "
          "also supported."),
    require_value=True
)
findclusters_ns.add_parameter(
    "version",
    help="Select clusters with specified Bright version."
)
findclusters_ns.add_parameter(
    "distro",
    help="Select clusters with specified base distribution."
)
findclusters_ns.add_enumeration_parameter(
    "tags",
    default=[],
    help="List of tags, the cluster(s) must have them all. (AND)"
)
findclusters_ns.add_enumeration_parameter(
    "tags_any",
    default=[],
    help="List of tags, the cluster(s) must have at least one of them. (OR)"
)
findclusters_ns.add_enumeration_parameter(
    "not_tags",
    default=[],
    help="List of tags, the cluster(s) must not have all of them. (NAND)"
)
findclusters_ns.add_enumeration_parameter(
    "not_tags_any",
    default=[],
    help="List of tags, the cluster(s) must not have any of them. (NOR)"
)
findclusters_ns.add_switch_parameter(
    "ignore_newer",
    help=("Ignore cluster which were created with a newer version of the "
          "cluster-on-demand tools. When such clusters are encountered, and "
          "this flag is not specified, the tools will abort with an error "
          "message.")
)
findclusters_ns.add_switch_parameter(
    "all_stacks",
    help=("Assume all heat stacks are clusters, even the ones that don't "
          "have the tags normally required. Use with care, especially if heat "
          "is also used for workloads other than cluster on demand.")
)



# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import, print_function

from enum import Enum, auto

from dateutil.tz import tzlocal
from prettytable import PrettyTable

from clusterondemand.utils import get_time_ago

SEPARATOR = "##SEP##"


class SummaryType(Enum):
    Proposal = auto()
    Overview = auto()


class SummaryGenerator(object):
    """Generate the summary for creation of clusters and nodes."""

    def __init__(self,
                 head_name,
                 config=None,
                 head_image=None,
                 ip=None,
                 nas_node_definition=None,
                 node_definitions=None,
                 node_image=None,
                 primary_head_node_definition=None,
                 secondary_head_node_definition=None,
                 region=None,
                 summary_type=None,
                 ssh_string=None,
                 trace_id=None):
        self._config = config
        self._head_name = head_name
        self._head_image = head_image
        self._ip = ip
        self._nas_node_definition = nas_node_definition
        self._node_definitions = node_definitions if node_definitions else []
        self._node_image = node_image
        self._primary_head_node_definition = primary_head_node_definition
        self._secondary_head_node_defintion = secondary_head_node_definition
        self._region = region
        self._type = summary_type
        self._ssh_string = ssh_string
        self._trace_id = trace_id

    def generate_table(self):
        """Generate a list of lines containing the cluster overview."""
        table = PrettyTable(header=False, border=False)
        table.field_names = ["Property", "Value"]
        table.align["Property"] = "r"
        table.align["Value"] = "l"

        self._add_header(table)
        if self._type == SummaryType.Proposal:
            self._add_image_rows(table)
            self._add_head_node_info(table)
            self._add_compute_node_info(table)
            self._add_nas_node_info_row(table)
        self._add_rows(table)
        if self._type == SummaryType.Overview:
            self._add_access_info(table)
        self._add_separator(table)
        table_lines = str(table).split("\n")

        # Replace each separator line with a row of dashes.
        separator_line = "-" * len(table_lines[0])
        lines = [separator_line if SEPARATOR in line else line for line in table_lines]

        return lines

    def print_summary(self, println=print):
        """
        Print the summary.

        :param println: Function to be used to print each line
        """
        for line in self.generate_table():
            println(line)

    def _add_compute_node_info(self, table):
        for definition in self._node_definitions:
            self._add_node_summary(table, "Nodes:", definition)

    def _add_cluster_name(self, table):
        table.add_row(["Cluster:", self._head_name])

    def _add_access_info(self, table):
        if self._ssh_string:
            table.add_row(["SSH string:", self._ssh_string])
        if self._ip:
            table.add_row(["IP:", self._ip])
        if self._config["ssh_pub_key_path"]:
            table.add_row(["Key path:", self._config["ssh_pub_key_path"]])
        if self._config["ssh_password_authentication"]:
            table.add_row(["Warning:", "SSH password authentication enabled"])
        if self._config["log_cluster_password"]:
            table.add_row(["Password:", self._config["cluster_password"]])

    def _add_rows(self, table):
        """
        Add COD specific infomation to the summary table.
        """
        pass

    def _add_header(self, table):
        self._add_separator(table)
        self._add_cluster_name(table)
        self._add_separator(table)

    def _add_head_node_info(self, table):
        if self._primary_head_node_definition and self._secondary_head_node_defintion:
            self._add_node_summary(table, "Primary head node:", self._primary_head_node_definition)
            self._add_node_summary(table, "Secondary head node:", self._secondary_head_node_defintion)
        elif self._primary_head_node_definition:
            self._add_node_summary(table, "Head node:", self._primary_head_node_definition)

    def _add_image_rows(self, table):
        if self._head_image and self._head_image.name:
            revision = f":{self._head_image.revision}" if self._head_image.revision else ""
            image_details = f"({self._head_image.id}{revision})" if self._head_image.id else ""
            table.add_row(["Image name:", f"{self._head_image.name}{image_details}"])

        if self._node_image and self._node_image.name:
            table.add_row(["Node image:", "%s(%s:%s)" % (
                self._node_image.name,
                self._node_image.id,
                self._node_image.revision,
            )])

        if self._head_image and self._head_image.name:
            img_created = self._head_image.created_at.astimezone(tzlocal()).strftime(format="%Y-%m-%d %H:%M")
            img_age = get_time_ago(self._head_image.created_at)
            pkg_groups = self._head_image.package_groups or "none"

            table.add_row(["Image date:", "%s (%s ago)" % (img_created, img_age)])
            table.add_row(["Package groups:", pkg_groups])

            if self._config and "show_git_hashes" in self._config and self._config["show_git_hashes"]:
                git_hashes = "\n".join(f"{pkg}: {git_hash}" for pkg, git_hash in self._head_image.pkg_hashes) or "n/a"
                table.add_row(["Git hashes:", git_hashes])

        if self._head_image and self._head_image.version:
            table.add_row(["Version:", self._head_image.version])

        if self._head_image and self._head_image.distro:
            table.add_row(["Distro:", self._head_image.distro])

    def _add_nas_node_info_row(self, table):
        if self._nas_node_definition:
            self._add_node_summary(table, "NAS node:", self._nas_node_definition)

    def _add_node_summary(self, table, field_name, node_definition):
        flavor_string = self._get_flavor_string(node_definition.flavor)
        summary = f"{node_definition.count} ({flavor_string})"

        # If the node count is set to 0 we don't want to add unnecessary information and can exit early.
        if node_definition.count == 0:
            table.add_row([field_name, summary])
            return

        if node_definition.availability_zone:
            summary += f"\nAvailability zone: {node_definition.availability_zone}"

        # The first disk is the root disk, all other disks are extra.
        root_disk = node_definition.disks[0] if node_definition.disks else None
        extra_disks = node_definition.disks[1:]

        if root_disk:
            root_disk_type = f""" (type: {root_disk.get("type")}) """ if root_disk.get("type") else ""
            summary += f"""\nRoot disk: {root_disk.get("size")} GiB{root_disk_type}"""

        for disk in extra_disks:
            if not disk.get("size"):
                continue
            extra_disk_type = f""" (type: {disk.get("type")}) """ if disk.get("type") else ""
            summary += f"""\nExtra disk: {disk.get("size")} GiB{extra_disk_type}"""

        table.add_row([field_name, summary])

    def _add_region(self, table):
        table.add_row(["Region:", self._region])

    def _add_separator(self, table):
        table.add_row([SEPARATOR, ""])

    def _add_trace_id(self, table):
        if self._trace_id:
            table.add_row(["Trace ID:", self._trace_id])

    def _get_flavor_string(self, flavor):
        return str(flavor)



# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Callable, TypeVar

RT = TypeVar("RT")


def static_vars(**kwargs: Any) -> Callable[[Callable[..., RT]], Callable[..., RT]]:
    """Set some local static variables in a function.

    This will be a global variable, but linked to a function by being an attribute of it

    Example:
    @static_vars(my_var=0):
    def f():
      f.my_var += 1
    """
    def decorator(func: Callable[..., RT]) -> Callable[..., RT]:
        for key, value in kwargs.items():
            setattr(func, key, value)
        return func
    return decorator



# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

from clusterondemand.exceptions import CODException
from clusterondemandconfig import ConfigNamespace, config

clusterprefix_ns = ConfigNamespace("cluster.prefix", help_section="Cluster name prefix")
clusterprefix_ns.add_parameter(
    "cod_prefix",
    advanced=True,
    env="COD_PREFIX",
    help="A prefix that is by default prepended to the cluster name when creating a new cluster"
)
clusterprefix_ns.add_switch_parameter(
    "require_cod_prefix",
    advanced=True,
    help=("Whether the tool should require prefix to be set. "
          "The default of this value is typically configured by the cloud-admin.")
)
clusterprefix_ns.add_switch_parameter(
    "enforce_lowercase_cod_prefix",
    advanced=True,
    help="Whether the prefix may only consist of lowercase characters."
)


def get_prefix(config=config):
    prefix = config["cod_prefix"]
    required = config["require_cod_prefix"]
    lowercase = config["enforce_lowercase_cod_prefix"]

    if not prefix and required:
        raise CODException("Please set COD_PREFIX in your environment.")
    elif not prefix:
        return ""
    elif lowercase and not (prefix.isalpha() and prefix.islower()):
        raise CODException("COD_PREFIX may only consist of lowercase alphabetic characters.")

    return prefix


def ensure_cod_prefix(name):
    """Prefix name with COD_PREFIX if needed.

    'name' may contain the prefix already, then 'name' is returned
    If 'name' is None, None is returned. This is useful in the case of an optional parameter
    """
    if name is None:
        return None

    return _ensure_string_starts_with_cod_prefix(name, get_prefix(config))


def must_start_with_cod_prefix(parameter, config):
    """A config validation that will ensure that the parameter value starts with cod_prefix."""
    cod_prefix = get_prefix(config)

    if parameter.type not in [str, [str]]:
        # TODO: CM-20585, handle this statically.
        raise CODException("This validation cannot be used with a parameter of type %s" % (parameter.type.__name__))

    if not cod_prefix or config[parameter.key] is None:
        return
    elif str == parameter.type:
        config[parameter.key] = _ensure_string_starts_with_cod_prefix(config[parameter.key], cod_prefix)
    elif [str] == parameter.type:
        config[parameter.key] = [
            _ensure_string_starts_with_cod_prefix(name, cod_prefix)
            for name in config[parameter.key]
        ]
    else:
        assert False


def _ensure_string_starts_with_cod_prefix(string, cod_prefix):
    if cod_prefix and not string.startswith(cod_prefix + "-"):
        return cod_prefix + "-" + string
    return string



# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import, print_function

import logging
import os
import sys
import textwrap

from six.moves import input

PATH_TO_CURRENT_EULA = os.path.join(os.path.dirname(os.path.realpath(__file__)), "eula.txt")

EULA_DIALOG_TEXT = ("""
Before this setup can continue, the End-User License Agreement for Bright CM must be read and
accepted. It can be found at:
  %s

Please exit this program with Ctrl+C and read the terms within this file. If this has already been
done, continue by typing 'accept' and pressing enter.

Once accepted, this dialog will not be shown again until a newer version of the EULA becomes available.
""" % (PATH_TO_CURRENT_EULA)).strip()

CANNOT_PROMPT_ACCEPT_ERROR = ("""
Cannot continue with this setup until the EULA has been accepted.

Either run this program with the flag --accept-eula, set the environment variable COD_ACCEPT_EULA=True,
or store the following in a cluster-on-demand config file:

----
[common.eula]
accept_eula=True
----
""")

log = logging.getLogger("cluster-on-demand")


def user_accepts_eula(configuration):
    """Return True when the user accepts, or previously accepted, the current EULA.

    When the user accepts the EULA, a copy of the text is written to a certain file
    (config["user_accepted_eula_file"]). As long as the contents of that file are identical to the
    contents of PATH_TO_CURRENT_EULA, we don't ask the user to accept the EULA again.
    """
    current_eula = _contents_of_file(PATH_TO_CURRENT_EULA)

    if _user_previously_accepted_current_eula(configuration, current_eula):
        return True
    elif _user_accepts_current_eula(configuration):
        _save_current_eula(configuration, current_eula)
        return True
    else:
        return False


def _user_previously_accepted_current_eula(configuration, current_eula):
    return os.path.exists(configuration["user_accepted_eula_file"]) and \
        _contents_of_file(configuration["user_accepted_eula_file"]) == current_eula


def _contents_of_file(path):
    with open(path, "r") as f:
        return f.read()


def _save_current_eula(configuration, current_eula):
    with open(configuration["user_accepted_eula_file"], "w") as f:
        f.write(current_eula)


def _user_accepts_current_eula(configuration):
    if configuration["accept_eula"]:
        log.debug("User accepted the EULA through the configuration.")
        return True

    if not sys.stdout.isatty():
        raise RuntimeError(CANNOT_PROMPT_ACCEPT_ERROR)

    print(textwrap.dedent(EULA_DIALOG_TEXT))

    while True:
        try:
            if _prompt_user_for_accept():
                print("The EULA was accepted. A copy is stored in %s" % (configuration["user_accepted_eula_file"]))
                return True
            else:
                print("Please either type 'accept' and press Enter, or exit with Ctrl+C")
        except KeyboardInterrupt:
            log.debug("User did not accept EULA, sent Ctrl+C instead")
            return False


def _prompt_user_for_accept():
    return "accept" == input()



# Copyright 2004-2023 Bright Computing Holding BV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import, division

from datetime import datetime

from dateutil import tz

DATE_TIME_FORMAT = "%Y-%m-%d %H:%M"


def format_datetime(date_time, format=DATE_TIME_FORMAT, tzinfo=None):
    """Format datetime in locale timezone.

    >>> import pytz
    >>> format_datetime(datetime(2009, 12, 6, 16, 29, tzinfo=pytz.UTC),
    ...                 tzinfo=pytz.timezone('Etc/GMT+1'))
    '2009-12-06 15:29'
    """
    tzinfo = tzinfo or tz.tzlocal()
    return date_time.astimezone(tzinfo).strftime(format)


def get_datetime_ago(then_datetime, now_datetime=None):
    """
    Return human-readable difference between two datetimes.

    :param now_datetime:
    :param then_datetime:
    :return:
    >>> import pytz
    >>> get_datetime_ago(
    ...     datetime(2006, 11, 5, 17, 29, 43, tzinfo=pytz.UTC),
    ...    datetime(2009, 12, 6, 16, 29, 43, tzinfo=pytz.UTC)
    ... )
    '1126d 23h'

    >>> get_datetime_ago(
    ... datetime(2009, 12, 6, 15, 2, 4, tzinfo=pytz.UTC),
    ... datetime(2009, 12, 6, 16, 29, 43, tzinfo=pytz.UTC)
    ... )
    '1h 27m'
    """
    now_datetime = now_datetime or datetime.now(then_datetime.tzinfo)

    delta = now_datetime - then_datetime
    days = delta.days
    minutes = delta.seconds // 60
    hours = minutes // 60
    minutes %= 60
    if days:
        return "%dd %dh" % (days, hours)
    else:
        return "%dh %dm" % (hours, minutes)
